# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import Callable, Dict, List, Optional, Set, Union

import numpy as np
import torch
from torch.utils.data import DataLoader

from nemo.collections.asr.parts.utils.aligner_utils import (
    BLANK_TOKEN,
    Segment,
    Word,
    add_t_start_end_to_utt_obj,
    get_batch_variables,
    viterbi_decoding,
)
from nemo.collections.asr.parts.utils.rnnt_utils import Hypothesis
from nemo.utils import logging, logging_mode


def flatten_char_offsets(char_offsets: List[Dict[str, Union[int, float]]]) -> List[Dict[str, Union[int, float]]]:
    """
    Flatten the char offsets to contain only one char and one token per offset.
    This is needed for RNNT decoding, as they return a list of strings for offset['char'].
    """
    if not char_offsets:
        return char_offsets

    flattened_char_offsets = []
    for char_offset in char_offsets:
        if isinstance(char_offset['char'], list):
            for char in char_offset['char']:
                sub_char_offset = char_offset.copy()
                sub_char_offset['char'] = char
                flattened_char_offsets.append(sub_char_offset)
        else:
            flattened_char_offsets.append(char_offset)
    return flattened_char_offsets


def get_words_offsets(
    char_offsets: List[Dict[str, Union[int, float]]],
    encoded_char_offsets: List[Dict[str, Union[int, float]]],
    decode_tokens_to_str: Callable[[List[int]], str],
    word_delimiter_char: str = " ",
    tokenizer_type: str = "bpe",
    supported_punctuation: Optional[Set] = None,
) -> List[Dict[str, Union[int, float]]]:
    """
    Utility method which constructs word time stamps out of sub-word time stamps.

    Args:
        char_offsets: A list of dictionaries, each containing "char", "start_offset" and "end_offset",
                    where "char" is decoded with the tokenizer.
        encoded_char_offsets: A list of dictionaries, each containing "char", "start_offset" and "end_offset",
                    where "char" is the original id/ids from the hypotheses (not decoded with the tokenizer).
                    This is needed for subword tokenization models.
        word_delimiter_char: Character token that represents the word delimiter. By default, " ".
        supported_punctuation: Set containing punctuation marks in the vocabulary.

    Returns:
        A list of dictionaries containing the word offsets. Each item contains "word", "start_offset" and
        "end_offset".
    """

    def define_word_start_condition() -> Callable[[str, str], bool]:
        """
        Define the word start condition based on the tokenizer type and word delimiter character.
        """
        if tokenizer_type in ["bpe", "wpe"] and word_delimiter_char == " ":
            if tokenizer_type == "wpe":
                return (
                    lambda token, token_text, next_non_delimeter_token: token_text
                    and not token_text.startswith("##")
                    or (token_text == word_delimiter_char and next_non_delimeter_token not in supported_punctuation)
                )
            return lambda token, token_text, next_non_delimeter_token: token != token_text or (
                token_text == word_delimiter_char and next_non_delimeter_token not in supported_punctuation
            )
        elif word_delimiter_char == " ":
            return (
                lambda token, token_text, next_non_delimeter_token: token_text == word_delimiter_char
                and next_non_delimeter_token not in supported_punctuation
            )
        else:
            return lambda token, token_text, next_non_delimeter_token: token_text == word_delimiter_char

    char_offsets = flatten_char_offsets(char_offsets)
    encoded_char_offsets = flatten_char_offsets(encoded_char_offsets)

    if encoded_char_offsets is None:
        encoded_char_offsets = char_offsets

    word_offsets = []
    previous_token_index = 0

    # Built tokens should be list here as when dealing with wpe tokenizer,
    # ids should be decoded together to ensure tokens starting with ## are not split
    built_tokens = []
    condition_for_word_start = define_word_start_condition()

    # For every collapsed sub-word token
    for i, (char_offset, char_token_offset) in enumerate(zip(char_offsets, encoded_char_offsets)):

        char_text = char_offset['char']
        char_token = char_token_offset['char']

        curr_punctuation = (
            supported_punctuation and char_text in supported_punctuation and char_text != word_delimiter_char
        )
        next_non_delimeter_token = None
        next_non_delimeter_token_index = i
        while not next_non_delimeter_token and next_non_delimeter_token_index < len(char_offsets) - 1:
            next_non_delimeter_token_index += 1
            next_non_delimeter_token = char_offsets[next_non_delimeter_token_index]['char']
            next_non_delimeter_token = (
                next_non_delimeter_token if next_non_delimeter_token != word_delimiter_char else None
            )
        # It is a sub-word token, or contains an identifier at the beginning such as _ or ## that was stripped
        # after forcing partial text conversion of the token.
        # AND it is not a supported punctuation mark, which needs to be added to the built word regardless of its identifier.

        if condition_for_word_start(char_token, char_text, next_non_delimeter_token) and not curr_punctuation:
            # If there are any partially or fully built sub-word token ids, construct to text.
            # Note: This is "old" subword, that occurs *after* current sub-word has started.

            if built_tokens:
                word_offsets.append(
                    {
                        "word": decode_tokens_to_str(built_tokens),
                        "start_offset": char_offsets[previous_token_index]["start_offset"],
                        "end_offset": char_offsets[i - 1]["end_offset"],
                    }
                )

                if "start" in char_offset:
                    word_offsets[-1]["start"] = char_offsets[previous_token_index]["start"]
                if "end" in char_offset:
                    word_offsets[-1]["end"] = char_offsets[i - 1]["end"]

            # Prepare new built_tokens
            built_tokens = []

            if char_text != word_delimiter_char:
                built_tokens.append(char_token)
                previous_token_index = i

        # If the token is a punctuation mark and there is no built word, then the previous word is complete
        # and lacks the punctuation mark. We need to add the punctuation mark to the previous formed word.
        elif curr_punctuation and not built_tokens and word_offsets:
            last_built_word = word_offsets[-1]
            last_built_word['end_offset'] = char_offset['end_offset']
            if last_built_word['word'][-1] == ' ':
                last_built_word['word'] = last_built_word['word'][:-1]
            last_built_word['word'] += char_text
        # If the token is a punctuation mark and there is a built word,
        # then we need to add the punctuation mark to the built word and remove preceding space.
        elif curr_punctuation and built_tokens:
            if built_tokens[-1] in [' ', "_", "▁"]:
                built_tokens = built_tokens[:-1]
            built_tokens.append(char_token)
        else:
            # If the token does not contain any sub-word start mark, then the sub-word has not completed yet
            # Append to current built word.
            # If this token is the first in the built_tokens, we should save its index as the previous token index
            # because it will be used to calculate the start offset of the word.
            if not built_tokens:
                previous_token_index = i
            built_tokens.append(char_token)

    # Inject the start offset of the first token to word offsets
    # This is because we always skip the delay the injection of the first sub-word due to the loop
    # condition and check whether built token is ready or not.
    # Therefore without this forced injection, the start_offset appears as off by 1.
    if len(word_offsets) == 0:
        # alaptev: sometimes word_offsets can be empty
        if built_tokens:
            word_offsets.append(
                {
                    "word": decode_tokens_to_str(built_tokens),
                    "start_offset": char_offsets[0]["start_offset"],
                    "end_offset": char_offsets[-1]["end_offset"],
                }
            )
            if "start" in char_offsets[0]:
                word_offsets[0]["start"] = char_offsets[0]["start"]
            if "end" in char_offsets[-1]:
                word_offsets[-1]["end"] = char_offsets[-1]["end"]
    else:
        word_offsets[0]["start_offset"] = char_offsets[0]["start_offset"]

        if "start" in char_offsets[0]:
            word_offsets[0]["start"] = char_offsets[0]["start"]

        # If there are any remaining tokens left, inject them all into the final word offset.
        # Note: The start offset of this token is the start time of the first token inside build_token.
        # Note: The end offset of this token is the end time of the last token inside build_token
        if built_tokens:
            word_offsets.append(
                {
                    "word": decode_tokens_to_str(built_tokens),
                    "start_offset": char_offsets[previous_token_index]["start_offset"],
                    "end_offset": char_offsets[-1]["end_offset"],
                }
            )
            if "start" in char_offset:
                word_offsets[-1]["start"] = char_offsets[previous_token_index]["start"]
            if "end" in char_offset:
                word_offsets[-1]["end"] = char_offsets[-1]["end"]

    return word_offsets


def get_segment_offsets(
    word_offsets: List[Dict[str, Union[str, float]]],
    segment_delimiter_tokens: List[str],
    supported_punctuation: Optional[Set] = None,
    segment_gap_threshold: Optional[int] = None,
) -> List[Dict[str, Union[str, float]]]:
    """
    Utility method which constructs segment time stamps out of word time stamps.

    Args:
        offsets: A list of dictionaries, each containing "word", "start_offset" and "end_offset".
        segments_delimiter_tokens: List containing tokens representing the seperator(s) between segments.
        supported_punctuation: Set containing punctuation marks in the vocabulary.
        segment_gap_threshold: Number of frames between 2 consecutive words necessary to form segments out of plain text.

    Returns:
        A list of dictionaries containing the segment offsets. Each item contains "segment", "start_offset" and
        "end_offset".
    """
    if (
        supported_punctuation
        and not set(segment_delimiter_tokens).intersection(supported_punctuation)
        and not segment_gap_threshold
    ):
        logging.warning(
            f"Specified segment seperators are not in supported punctuation {supported_punctuation}. "
            "If the seperators are not punctuation marks, ignore this warning. "
            "Otherwise, specify 'segment_gap_threshold' parameter in decoding config to form segments.",
            mode=logging_mode.ONCE,
        )

    segment_offsets = []
    segment_words = []
    previous_word_index = 0

    # For every offset word
    for i, offset in enumerate(word_offsets):

        word = offset['word']
        if segment_gap_threshold and segment_words:
            gap_between_words = offset['start_offset'] - word_offsets[i - 1]['end_offset']

            if gap_between_words >= segment_gap_threshold:
                segment_offsets.append(
                    {
                        "segment": ' '.join(segment_words),
                        "start_offset": word_offsets[previous_word_index]["start_offset"],
                        "end_offset": word_offsets[i - 1]["end_offset"],
                    }
                )

                if "start" in word_offsets[previous_word_index]:
                    segment_offsets[-1]["start"] = word_offsets[previous_word_index]["start"]
                if "end" in word_offsets[i - 1]:
                    segment_offsets[-1]["end"] = word_offsets[i - 1]["end"]

                segment_words = [word]
                previous_word_index = i
                continue

        # check if the word ends with any delimeter token or the word itself is a delimeter
        elif word and (word[-1] in segment_delimiter_tokens or word in segment_delimiter_tokens):
            segment_words.append(word)
            if segment_words:
                segment_offsets.append(
                    {
                        "segment": ' '.join(segment_words),
                        "start_offset": word_offsets[previous_word_index]["start_offset"],
                        "end_offset": offset["end_offset"],
                    }
                )

                if "start" in word_offsets[previous_word_index]:
                    segment_offsets[-1]["start"] = word_offsets[previous_word_index]["start"]
                if "end" in offset:
                    segment_offsets[-1]["end"] = offset["end"]

            segment_words = []
            previous_word_index = i + 1
            continue

        segment_words.append(word)

    if segment_words:
        start_offset = word_offsets[previous_word_index]["start_offset"]
        segment_offsets.append(
            {
                "segment": ' '.join(segment_words),
                "start_offset": start_offset,
                "end_offset": word_offsets[-1]["end_offset"],
            }
        )

        if "start" in word_offsets[previous_word_index]:
            segment_offsets[-1]["start"] = word_offsets[previous_word_index]["start"]
        if "end" in word_offsets[-1]:
            segment_offsets[-1]["end"] = word_offsets[-1]["end"]

    segment_words.clear()

    return segment_offsets


def process_aed_timestamp_outputs(outputs, subsampling_factor: int = 1, window_stride: float = 0.01):
    """
    Processes AED timestamp outputs and extracts word-level timestamps.
    Args:
        outputs (Hypothesis, list of Hypotesis or list of list of Hypotesis): The hypothesis outputs to process. Can be a single Hypothesis object or a list of Hypothesis objects.
        subsampling_factor (int, optional): The subsampling factor used in the model. Default is 1.
        window_stride (float, optional): The window stride used in the model. Default is 0.01.
    Returns:
        list of list of Hypotesis: The processed hypothesis outputs with word-level timestamps added.
    """

    def extract_words_with_timestamps(text, subsampling_factor: int = 1, window_stride: float = 0.01):
        text = text.strip()  # remove leading and trailing whitespaces - training data artifact

        if not re.search(r'<\|\d+\|>.*?<\|\d+\|>', text):
            return None, text

        # Find words that directly have start and end timestamps
        pattern = r'<\|(\d+)\|>(.*?)<\|(\d+)\|>'

        matches = []
        text_without_timestamps = []
        for match in re.finditer(pattern, text):
            start_offset = int(match.group(1))
            content = match.group(2)
            end_offset = int(match.group(3))
            start_time = start_offset * window_stride * subsampling_factor
            end_time = end_offset * window_stride * subsampling_factor

            # Only include if there's actual content
            if content.strip():
                sample = {
                    'word': content.strip(),
                    'start_offset': start_offset,
                    'end_offset': end_offset,
                    'start': start_time,
                    'end': end_time,
                }
                matches.append(sample)
                text_without_timestamps.append(content.strip())

        text_without_timestamps = ' '.join(text_without_timestamps)
        return matches, text_without_timestamps

    def segments_offset_to_time(segments, window_stride, subsampling_factor):
        for segment in segments:
            segment['start'] = segment['start_offset'] * window_stride * subsampling_factor
            segment['end'] = segment['end_offset'] * window_stride * subsampling_factor
        return segments

    def process_hypothesis(hyp, subsampling_factor: int, window_stride: float):
        """
        Processes a single Hypothesis object to extract timestamps.
        """
        timestamp, text = extract_words_with_timestamps(hyp.text, subsampling_factor, window_stride)
        hyp.text = text
        if timestamp is not None:
            if len(hyp.timestamp) == 0:
                hyp.timestamp = {}

            hyp.timestamp.update(
                {
                    'word': timestamp,
                    'segment': [],
                    'char': [],  # not supported for AED
                }
            )

            segments = get_segment_offsets(timestamp, segment_delimiter_tokens=['.', '?', '!'])
            hyp.timestamp['segment'] = segments_offset_to_time(segments, window_stride, subsampling_factor)
        else:
            hyp.timestamp = {
                'word': [],
                'segment': [],
                'char': [],
            }

        return hyp

    if outputs is None:
        return outputs

    if isinstance(outputs, Hypothesis):
        return [process_hypothesis(outputs, subsampling_factor, window_stride)]
    elif isinstance(outputs, list) and isinstance(outputs[0], Hypothesis):
        # list of Hypothesis
        return [process_hypothesis(hyp, subsampling_factor, window_stride) for hyp in outputs]
    elif isinstance(outputs, list) and isinstance(outputs[0], list) and isinstance(outputs[0][0], Hypothesis):
        # list of list of Hypothesis (for beam decoding)
        return [
            [process_hypothesis(hyp, subsampling_factor, window_stride) for hyp in hyps_list] for hyps_list in outputs
        ]
    else:
        raise ValueError(
            f"Expected Hypothesis, list of Hypothesis or list of list of Hypothesis object, got {type(outputs)}"
        )


def process_timestamp_outputs(outputs, subsampling_factor: int = 1, window_stride: float = 0.01):
    """
    Process the timestamps from list of hypothesis to user friendly format.
    Converts the start and end duration from frames to seconds.
    Args:
        outputs: List of Hypothesis objects.
        subsampling_factor: int, Subsampling factor used in the model.
        window_stride: float, Window stride used in the model. (sometimes referred to as hop length/shift)
    Returns:
        List of Hypothesis objects with processed timestamps
    """

    if outputs is None:
        return outputs

    if isinstance(outputs, Hypothesis):
        outputs = [outputs]

    if not isinstance(outputs[0], Hypothesis):
        raise ValueError(f"Expected Hypothesis object, got {type(outputs[0])}")

    def process_timestamp(timestamp, subsampling_factor, window_stride):
        """
        Process the timestamp for a single hypothesis.
        return the start and end duration in seconds.
        """
        for idx, val in enumerate(timestamp):
            start_offset = val['start_offset']
            end_offset = val['end_offset']
            start = start_offset * window_stride * subsampling_factor
            end = end_offset * window_stride * subsampling_factor
            val['start'] = start
            val['end'] = end

        return timestamp

    for idx, hyp in enumerate(outputs):
        if not hasattr(hyp, 'timestamp'):
            raise ValueError(
                f"Expected Hypothesis object to have 'timestamp' attribute, when compute_timestamps is \
                    enabled but got {hyp}"
            )
        timestamp = hyp.timestamp
        if 'word' in timestamp:
            outputs[idx].timestamp['word'] = process_timestamp(timestamp['word'], subsampling_factor, window_stride)
        if 'char' in timestamp:
            outputs[idx].timestamp['char'] = process_timestamp(timestamp['char'], subsampling_factor, window_stride)
        if 'segment' in timestamp:
            outputs[idx].timestamp['segment'] = process_timestamp(
                timestamp['segment'], subsampling_factor, window_stride
            )
    return outputs


def get_forced_aligned_timestamps_with_external_model(
    audio: Union[str, List[str], np.ndarray, DataLoader],
    external_ctc_model,
    main_model_predictions: List[Hypothesis],
    batch_size: int = 4,
    viterbi_device: Optional[torch.device] = None,
    segment_separators: Optional[Union[str, List[str]]] = ['.', '?', '!', '...'],
    word_separator: Optional[str] = " ",
    supported_punctuation: Optional[Union[Set, List[str]]] = {',', '.', '!', '?'},
    timestamp_type: Optional[Union[str, List[str]]] = "all",
    has_hypotheses: bool = False,
) -> List[Hypothesis]:
    """
    Extracts the word, segment and char timestamps by aligning the audio with the external ASR model and adds them to the provided Hypothesis objects.
    Args:
        audio: The audio to align.
        external_ctc_model: The external ASR CTC model to use for alignment.
        main_model_predictions: The predictions from the main model the pred_texts of which will be used for alignment.
        batch_size: The batch size to use for alignment (this is used both for CTC model inference and viterbi decoding).
        viterbi_device: The device to use for viterbi decoding. Batch variables got with get_batch_variables() are moved to this device before viterbi decoding.
        segment_separators: The segment separators to use for splitting the pred_text into segments. Default is ['.', '?', '!', '...']
        word_separator: The word separator to use for splitting the pred_text into words. Default is " ".
        supported_punctuation: The supported punctuation is punctuation marks in the vocabulary of the main model.
                            This is used for refining the timestamps extracted with the external ASR model.
                            As sometimes punctuation marks can be assigned to multiple audio frames, which is not correct, so we should neutralize these cases.
                            Default is {',', '.', '!', '?'}.
        timestamp_type: The type of timestamps to return. Default is "all". Can be "segment", "word", "char" or "all", or a list of these.
        has_hypotheses: Whether `audio` is a list of Hypothesis objects resulted from the external ASR CTC model inference.
                        This is used in external alignment generation script, e.g. `examples/asr/asr_chunked_inference/aed/speech_to_text_aed_chunked_infer.py`.
                        If True, `audio` will be used as a list of Hypothesis objects and the inference to the external ASR CTC model will be skipped.

    Returns:
        List of provided Hypothesis objects with processed timestamps
    """

    def process_timestamps(utt_obj, output_timestep_duration, timestamp_type):
        if isinstance(timestamp_type, str):
            assert timestamp_type in [
                "segment",
                "word",
                "char",
                "all",
            ], "Invalid timestamp type must be one of: segment, word, char, all"
            timestamp_type = [timestamp_type] if timestamp_type != "all" else ["segment", "word", "char"]
        elif isinstance(timestamp_type, list):
            assert all(
                t in ["segment", "word", "char", "all"] for t in timestamp_type
            ), "Invalid timestamp type must be one of: segment, word, char, all"
        else:
            raise ValueError("Invalid timestamp type must be one of: segment, word, char, all")

        timestamps = {t: [] for t in timestamp_type}

        for segment in utt_obj.segments_and_tokens:

            if not isinstance(segment, Segment):
                continue

            if "segment" in timestamp_type:
                timestamps["segment"].append(
                    {
                        "segment": segment.text,
                        "start_offset": (
                            int(segment.t_start / output_timestep_duration) if segment.t_start != -1 else -1
                        ),
                        "end_offset": int(segment.t_end / output_timestep_duration) if segment.t_end != -1 else -1,
                        "start": round(segment.t_start, 2),
                        "end": round(segment.t_end, 2),
                    }
                )

            for word in segment.words_and_tokens:

                if not isinstance(word, Word):
                    continue

                if "word" in timestamp_type:
                    timestamps["word"].append(
                        {
                            "word": word.text,
                            "start_offset": int(word.t_start / output_timestep_duration) if word.t_start != -1 else -1,
                            "end_offset": int(word.t_end / output_timestep_duration) if word.t_end != -1 else -1,
                            "start": round(word.t_start, 2),
                            "end": round(word.t_end, 2),
                        }
                    )

                for i, token in enumerate(word.tokens):
                    if token.text == BLANK_TOKEN:
                        continue

                    if token.text in supported_punctuation:

                        previous_non_blank_token_idx = i - 1
                        previous_non_blank_token = (
                            word.tokens[previous_non_blank_token_idx] if previous_non_blank_token_idx >= 0 else None
                        )

                        while previous_non_blank_token is not None and previous_non_blank_token.text == BLANK_TOKEN:
                            previous_non_blank_token_idx -= 1
                            previous_non_blank_token = (
                                word.tokens[previous_non_blank_token_idx]
                                if previous_non_blank_token_idx >= 0
                                else None
                            )

                        previous_token_end = (
                            round(previous_non_blank_token.t_end, 2)
                            if previous_non_blank_token is not None
                            else round(token.t_start, 2)
                        )

                        if "segment" in timestamp_type:
                            if segment.t_end == word.t_end:
                                timestamps["segment"][-1]["end"] = previous_token_end
                                timestamps["segment"][-1]["end_offset"] = (
                                    int(previous_token_end / output_timestep_duration)
                                    if previous_token_end != -1
                                    else -1
                                )

                        if "word" in timestamp_type:
                            if word.t_end == token.t_end:
                                timestamps["word"][-1]["end"] = previous_token_end
                                timestamps["word"][-1]["end_offset"] = (
                                    int(previous_token_end / output_timestep_duration)
                                    if previous_token_end != -1
                                    else -1
                                )

                        token.t_end = token.t_start = previous_token_end

                    if "char" in timestamp_type:
                        timestamps["char"].append(
                            {
                                "char": external_ctc_model.tokenizer.ids_to_text([token.token_id]),
                                "token_id": token.token_id,
                                "token": token.text,
                                "start_offset": (
                                    int(token.t_start / output_timestep_duration) if token.t_start != -1 else -1
                                ),
                                "end_offset": int(token.t_end / output_timestep_duration) if token.t_end != -1 else -1,
                                "start": round(token.t_start, 2),
                                "end": round(token.t_end, 2),
                            }
                        )

        return timestamps

    if viterbi_device is None:
        viterbi_device = external_ctc_model.device

    if timestamp_type == "char":
        segment_separators = None
        word_separator = None
    elif timestamp_type == "word":
        segment_separators = None

    for start_idx in range(0, len(audio), batch_size):
        end_idx = start_idx + batch_size

        audio_batch = audio[start_idx:end_idx]

        log_probs_batch, y_batch, T_batch, U_batch, utt_obj_batch, output_timestep_duration = get_batch_variables(
            audio=audio_batch,
            model=external_ctc_model,
            segment_separators=segment_separators,
            word_separator=word_separator,
            gt_text_batch=[hyp.text for hyp in main_model_predictions[start_idx:end_idx]],
            has_hypotheses=has_hypotheses,
        )

        alignments_batch = viterbi_decoding(
            log_probs_batch,
            y_batch,
            T_batch,
            U_batch,
            viterbi_device=viterbi_device,
        )

        for i, (utt_obj, alignment_utt) in enumerate(zip(utt_obj_batch, alignments_batch)):
            utt_obj = add_t_start_end_to_utt_obj(utt_obj, alignment_utt, output_timestep_duration)
            main_model_predictions[start_idx + i].timestamp = process_timestamps(
                utt_obj, output_timestep_duration, timestamp_type
            )

    return main_model_predictions



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest.mock import patch

import pytest

from nemo.collections.llm.recipes.run.executor import torchrun


@patch('torch.cuda.is_available')
@patch('torch.cuda.device_count')
def test_torchrun_with_explicit_devices(mock_device_count, mock_cuda_available):
    """Test torchrun factory with automatic device detection"""
    # Mock CUDA being available with 2 devices
    mock_cuda_available.return_value = True
    mock_device_count.return_value = 2

    config = torchrun(devices=2)

    assert config.ntasks_per_node == 2
    assert config.launcher == 'torchrun'
    assert config.env_vars == {'TORCH_NCCL_AVOID_RECORD_STREAMS': '1'}


@patch('torch.cuda.is_available')
def test_torchrun_raises_error_without_cuda(mock_cuda_available):
    """Test torchrun factory raises error when CUDA is not available and devices not specified"""
    mock_cuda_available.return_value = False

    with pytest.raises(RuntimeError) as exc_info:
        torchrun(devices=None)

    assert "Cannot infer the 'ntasks_per_node' parameter" in str(exc_info.value)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from pathlib import Path
from typing import Any, Dict, Optional, Union

from huggingface_hub import CONFIG_NAME, PyTorchModelHubMixin
from huggingface_hub.hub_mixin import DataclassInstance
from omegaconf import DictConfig, OmegaConf
from transformers.utils import cached_file


class HFHubMixin(
    PyTorchModelHubMixin,
    library_name="NeMo",
    repo_url="https://github.com/NVIDIA/NeMo",
    docs_url="https://docs.nvidia.com/nemo-framework/user-guide/latest/nemotoolkit",
):
    @classmethod
    def _from_pretrained(
        cls,
        *,
        model_id: str,
        revision: Optional[str],
        cache_dir: Optional[Union[str, Path]],
        force_download: bool,
        proxies: Optional[dict],
        resume_download: Optional[bool],
        local_files_only: bool,
        token: Union[str, bool, None],
        map_location: str = "cpu",
        strict: bool = False,
        **model_kwargs,
    ):
        """
        Load Pytorch pretrained weights and return the loaded model.
        Wrapper over PyTorchModelHubMixin that auto-handles config in **model_kwargs.
        """
        resolved_config_file = cached_file(
            model_id,
            CONFIG_NAME,
            cache_dir=cache_dir,
            force_download=force_download,
            resume_download=resume_download,
            proxies=proxies,
            local_files_only=local_files_only,
            token=token,
            revision=revision,
            _raise_exceptions_for_gated_repo=False,
            _raise_exceptions_for_missing_entries=False,
            _raise_exceptions_for_connection_errors=False,
        )
        if resolved_config_file is None:
            raise RuntimeError(f"Missing {CONFIG_NAME} file for {model_id=}")
        model_kwargs['cfg'] = OmegaConf.to_container(OmegaConf.load(resolved_config_file))
        # The setting below tells the model's __init__ not to load the original pretrained weights
        # for individual children modules.
        # To illustrate: if you trained a new model M using a pretrained ASR and a pretrained LLM,
        # this setting skips loading the original pretrained ASR and LLM weights, and loads the
        # final trained model weights directly.
        model_kwargs['cfg']['pretrained_weights'] = False
        return super()._from_pretrained(
            model_id=model_id,
            revision=revision,
            cache_dir=cache_dir,
            force_download=force_download,
            proxies=proxies,
            resume_download=resume_download,
            local_files_only=local_files_only,
            token=token,
            map_location=map_location,
            strict=strict,
            **model_kwargs,
        )

    def save_pretrained(
        self,
        save_directory: Union[str, Path],
        *,
        config: Optional[Union[dict, "DataclassInstance"]] = None,
        repo_id: Optional[str] = None,
        push_to_hub: bool = False,
        model_card_kwargs: Optional[Dict[str, Any]] = None,
        **push_to_hub_kwargs,
    ) -> Optional[str]:
        """
        Save weights in local directory.

        Args:
            save_directory (`str` or `Path`):
                Path to directory in which the model weights and configuration will be saved.
            config (`dict` or `DataclassInstance`, *optional*):
                Model configuration specified as a key/value dictionary or a dataclass instance.
                If not provided, we will automatically serialize attribute ``model.cfg``.
            push_to_hub (`bool`, *optional*, defaults to `False`):
                Whether or not to push your model to the Huggingface Hub after saving it.
            repo_id (`str`, *optional*):
                ID of your repository on the Hub. Used only if `push_to_hub=True`. Will default to the folder name if
                not provided.
            model_card_kwargs (`Dict[str, Any]`, *optional*):
                Additional arguments passed to the model card template to customize the model card.
            push_to_hub_kwargs:
                Additional key word arguments passed along to the [`~ModelHubMixin.push_to_hub`] method.
        Returns:
            `str` or `None`: url of the commit on the Hub if `push_to_hub=True`, `None` otherwise.
        """
        if config is None:
            config = getattr(self, "cfg")
            if isinstance(config, DictConfig):
                config = OmegaConf.to_container(self.cfg)
        return super().save_pretrained(
            save_directory=save_directory,
            config=config,
            repo_id=repo_id,
            push_to_hub=push_to_hub,
            model_card_kwargs=model_card_kwargs,
            **push_to_hub_kwargs,
        )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: LicenseRef-Apache2
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example Usage:
torchrun --nproc_per_node=2 tests/collections/llm/gpt/model/test_hyena_mixer_cp.py --operator_type hyena_short_conv [--use_subquadratic_ops]
"""

import argparse
import os
import time
from datetime import timedelta

import torch
import torch.distributed as dist
from einops import rearrange
from megatron.core import parallel_state
from megatron.core.tensor_parallel.random import model_parallel_cuda_manual_seed
from torch.distributed.nn.functional import all_gather as functional_all_gather
from torch.nn.parallel import DistributedDataParallel as DDP

from nemo.collections.llm.gpt.model.hyena import HyenaTestConfig
from nemo.collections.llm.gpt.model.megatron.hyena.hyena_config import HyenaConfig
from nemo.collections.llm.gpt.model.megatron.hyena.hyena_layer_specs import hyena_stack_spec_no_te
from nemo.collections.llm.gpt.model.megatron.hyena.hyena_mixer import HyenaMixer
from nemo.utils import logging


def init_parallel_state(tensor_model_parallel_size=1, pipeline_model_parallel_size=1, context_parallel_size=1):
    """Initialize distributed training and megatron parallel state."""

    num_gpus = torch.cuda.device_count()
    required_world_size = tensor_model_parallel_size * pipeline_model_parallel_size * context_parallel_size
    assert (
        num_gpus == required_world_size
    ), f"World size {num_gpus} != TP={tensor_model_parallel_size} x PP={pipeline_model_parallel_size} x CP={context_parallel_size}"

    # Set up environment variables
    os.environ["TORCH_NCCL_BLOCKING_WAIT"] = "0"
    os.environ["TORCH_NCCL_ASYNC_ERROR_HANDLING"] = "1"

    # Get local rank
    local_rank = int(os.getenv("LOCAL_RANK", 0))

    # Set device
    torch.cuda.set_device(local_rank)

    # Set up timeout
    timeout_seconds = int(os.getenv("TORCH_NCCL_HEARTBEAT_TIMEOUT_SEC", 1800))
    timeout_timedelta = timedelta(seconds=timeout_seconds)

    # Initialize process group if not already initialized
    if not dist.is_initialized():
        dist.init_process_group(backend="nccl", init_method="env://", timeout=timeout_timedelta)
        logging.info(f"Initialized distributed training with local rank {local_rank}")

    # Initialize parallel state
    parallel_state.initialize_model_parallel(
        tensor_model_parallel_size=tensor_model_parallel_size,
        pipeline_model_parallel_size=pipeline_model_parallel_size,
        context_parallel_size=context_parallel_size,
    )

    # Verify initialization
    cp_rank = parallel_state.get_context_parallel_rank()
    cp_world_size = parallel_state.get_context_parallel_world_size()
    logging.info(f"CP rank: {cp_rank}, CP world size: {cp_world_size}")
    return local_rank


def zigzag_split_across_group_ranks(data, group, seq_dim=0):
    """Distributes tensor data across group ranks using zigzag pattern.

    Divides the input tensor along sequence dimension and distributes chunks
    in an alternating pattern across different ranks.

    Arguments:
        data: original tensor to split across group ranks.
        group: the group to distribute the data across.
        seq_dim: the sequence/context dimension to split.

    Returns:
        Tensor slice for the current rank following zigzag distribution.
    """
    # Get group information
    process_count = len(dist.get_process_group_ranks(group))
    current_rank = dist.get_rank(group)

    # Skip distribution for single process
    if process_count == 1:
        return data

    # Calculate number of chunks for zigzag distribution
    total_chunks = 2 * process_count

    # Divide data into equal chunks
    tensor_chunks = list(torch.chunk(data, total_chunks, dim=seq_dim))

    # Implement zigzag distribution logic:
    # Each rank gets two chunks in specific positions
    # First chunk is at position equal to rank
    first_chunk_idx = current_rank
    # Second chunk is from the end, offset by rank+1
    second_chunk_idx = total_chunks - 1 - current_rank

    # Combine the appropriate chunks for this rank
    rank_data = torch.cat([tensor_chunks[first_chunk_idx], tensor_chunks[second_chunk_idx]], dim=seq_dim)

    return rank_data.contiguous()


def zigzag_gather_from_group_ranks(data, group, seq_dim=0):
    """Reconstructs complete tensor from zigzag-distributed chunks.

    Takes data distributed across ranks in zigzag pattern and reassembles
    the original complete tensor.

    Arguments:
        data: tensor fragment from current rank to be gathered.
        group: the group to gather data from.
        seq_dim: dimension along which to concatenate fragments.

    Returns:
        Reconstructed tensor with fragments from all ranks.
    """
    # Get group information
    process_count = len(dist.get_process_group_ranks(group))

    # Skip gathering for single process
    if process_count == 1:
        return data

    # Gather from all ranks using autograd-enabled all_gather
    gathered_data = functional_all_gather(data, group=group)

    # Initialize a list to store the original sequence chunks with proper tensor type
    seq_chunks = []
    for i in range(2 * process_count):
        seq_chunks.append(None)  # Will be replaced with tensors

    # Process each gathered tensor
    for i, data_i in enumerate(gathered_data):
        chunk_size = data_i.size(seq_dim) // 2

        # Split the data_i back into the original two chunks
        chunk0, chunk1 = torch.split(data_i, chunk_size, dim=seq_dim)

        # Reassign the chunks to their original positions
        seq_chunks[i] = chunk0
        seq_chunks[-(i + 1)] = chunk1

    # Concatenate all chunks to reconstruct the original data
    reconstructed_data = torch.cat(seq_chunks, dim=seq_dim)

    return reconstructed_data


class MixerModuleWrapper(torch.nn.Module):
    def __init__(self, seq_len, operator_type="hyena_short_conv", use_subquadratic_ops=False):
        super().__init__()

        self.use_subquadratic_ops = use_subquadratic_ops
        self.operator_type = operator_type

        # Create necessary submodules - use the mixer submodules like in the regular mixer fixture
        submodules = hyena_stack_spec_no_te.submodules.hyena_layer.submodules.mixer.submodules

        # Set the b2b parameter in the config
        hyena_config = HyenaConfig(num_groups_hyena=4096, num_groups_hyena_short=256, num_groups_hyena_medium=256)
        hyena_test_config = HyenaTestConfig(params_dtype=torch.float32, use_subquadratic_ops=use_subquadratic_ops)

        logging.info("Creating HyenaMixer...")
        self.mixer = HyenaMixer(
            transformer_config=hyena_test_config,
            hyena_config=hyena_config,
            max_sequence_length=seq_len,
            submodules=submodules,
            layer_number=1,
            operator_type=operator_type,
        )

    def forward(self, x, _use_cp=True):
        if self.use_subquadratic_ops and self.operator_type != "hyena":
            logging.info(f"Using subquadratic_ops: {self.use_subquadratic_ops}")
            z = self.mixer.b2b_kernel(x, _use_cp=_use_cp)
        else:
            logging.info("Using PyTorch implementation")
            features = self.mixer.hyena_proj_conv(x, _use_cp=_use_cp)
            x1, x2, v = rearrange(
                features, "b (g dg p) l -> b (g dg) p l", p=3, g=self.mixer.num_groups_per_tp_rank
            ).unbind(dim=2)
            z = self.mixer.mixer(x1, x2, v, _hyena_use_cp=_use_cp)
        return z


if __name__ == "__main__":
    # Parse command line arguments
    parser = argparse.ArgumentParser(description="Test hyena mixer with context parallelism")
    parser.add_argument(
        "--use_subquadratic_ops",
        action="store_true",
        default=False,
        help="Whether to use subquadratic_ops implementation",
    )
    parser.add_argument(
        "--operator_type",
        type=str,
        default="hyena_short_conv",
        choices=["hyena_short_conv", "hyena_medium_conv", "hyena"],
        help="Operator type. Options: hyena_short_conv, hyena_medium_conv, hyena",
    )
    parser.add_argument(
        "--tensor_model_parallel_size",
        type=int,
        default=1,
        help="Tensor model parallel size",
    )
    parser.add_argument(
        "--pipeline_model_parallel_size",
        type=int,
        default=1,
        help="Pipeline model parallel size",
    )
    parser.add_argument(
        "--context_parallel_size",
        type=int,
        default=2,
        help="Context parallel size",
    )
    parser.add_argument(
        "--log_dir",
        type=str,
        default="/tmp/nemo2_hyena_results",
        help="Directory for logs",
    )
    args = parser.parse_args()

    # Create log directory
    os.makedirs(args.log_dir, exist_ok=True)

    # Set up file handler for NeMo logging
    rank = int(os.getenv("RANK", "0"))
    log_file = os.path.join(args.log_dir, f"rank_{rank}.log")
    logging.add_file_handler(log_file)

    # Initialize parallel state
    local_rank = init_parallel_state(
        tensor_model_parallel_size=args.tensor_model_parallel_size,
        pipeline_model_parallel_size=args.pipeline_model_parallel_size,
        context_parallel_size=args.context_parallel_size,
    )

    logging.info(f"Starting hyena mixer test with args: {args}")

    try:
        # Initialize the model parallel RNG
        model_parallel_cuda_manual_seed(42)

        # Model initialization
        batch_size = 2
        seq_len = 1024  # Increased from 512 to provide more space for overlapping patches
        mixer_module_wrapper = MixerModuleWrapper(
            seq_len=seq_len,
            operator_type=args.operator_type,
            use_subquadratic_ops=args.use_subquadratic_ops,
        )

        ddp_mixer_module_wrapper = DDP(
            mixer_module_wrapper,
            process_group=parallel_state.get_data_parallel_group(with_context_parallel=True),
            find_unused_parameters=True,
        )

        input_features = torch.rand(
            (batch_size, mixer_module_wrapper.mixer.hidden_size * 3, seq_len),
            dtype=mixer_module_wrapper.mixer.transformer_config.params_dtype,
            device=torch.cuda.current_device(),
        )

        # Broadcast within each group
        cp_group = parallel_state.get_context_parallel_group()
        dist.broadcast(input_features, min(dist.get_process_group_ranks(cp_group)), group=cp_group)

        logging.info("Running without context parallel")
        output_features = ddp_mixer_module_wrapper(input_features, _use_cp=False)

        if dist.get_rank() == 0:
            try:
                assert output_features.shape == (
                    batch_size,
                    mixer_module_wrapper.mixer.hidden_size,
                    seq_len,
                ), f"output_features.shape: {output_features.shape}, batch_size: {batch_size}, mixer_module_wrapper.mixer.hidden_size: {mixer_module_wrapper.mixer.hidden_size}, seq_len: {seq_len}"
                logging.info(f"Output features shape: {output_features.shape}")
            except AssertionError as e:
                logging.error(f"Assertion error for output features shape: {e}")
                raise

        loss = output_features.float().mean()
        loss.backward()
        dist.barrier()

        # Store the gradients for later comparison.
        grads_without_cp = []
        for n, p in ddp_mixer_module_wrapper.named_parameters():
            if p.grad is not None:
                grads_without_cp.append((n, p.grad.clone()))

        ddp_mixer_module_wrapper.zero_grad()
        dist.barrier()

        logging.info("Running with context parallel")
        # Split the input features across the context parallel group
        input_features_cp = zigzag_split_across_group_ranks(input_features, group=cp_group, seq_dim=2)

        output_features_cp = ddp_mixer_module_wrapper(input_features_cp, _use_cp=True)
        if dist.get_rank() == 0:
            try:
                assert output_features_cp.shape == (
                    batch_size,
                    mixer_module_wrapper.mixer.hidden_size,
                    seq_len // parallel_state.get_context_parallel_world_size(),
                ), f"output_features_cp.shape: {output_features_cp.shape}, batch_size: {batch_size}, mixer_module_wrapper.mixer.hidden_size: {mixer_module_wrapper.mixer.hidden_size}, seq_len: {seq_len}"
                logging.info(f"Output features CP shape: {output_features_cp.shape}")
            except AssertionError as e:
                logging.error(f"Assertion error for output features CP shape: {e}")
                raise

        # Gather from all ranks according to zigzag splitting.
        output_features_cp_gathered = zigzag_gather_from_group_ranks(output_features_cp, group=cp_group, seq_dim=2)
        if dist.get_rank() == 0:
            try:
                # Verify shapes are correct
                assert output_features_cp_gathered.shape == (
                    batch_size,
                    mixer_module_wrapper.mixer.hidden_size,
                    seq_len,
                ), f"output_features_cp_gathered.shape: {output_features_cp_gathered.shape}, batch_size: {batch_size}, mixer_module_wrapper.mixer.hidden_size: {mixer_module_wrapper.mixer.hidden_size}, seq_len: {seq_len}"
                logging.info(f"Output features CP gathered shape: {output_features_cp_gathered.shape}")
            except AssertionError as e:
                logging.error(f"Assertion error for output features CP gathered shape: {e}")
                raise

        loss_with_cp = output_features_cp_gathered.float().mean()
        loss_with_cp.backward()
        dist.barrier()

        # Store the gradients for later comparison.
        grads_with_cp = []
        for n, p in ddp_mixer_module_wrapper.named_parameters():
            if p.grad is not None:
                grads_with_cp.append((n, p.grad.clone()))

        ddp_mixer_module_wrapper.zero_grad()
        dist.barrier()

        # Only perform comparison on rank 0
        if dist.get_rank() == 0:
            logging.info(f"Comparing loss values: without CP = {loss.item()}, with CP = {loss_with_cp.item()}")
            try:
                torch.testing.assert_close(loss, loss_with_cp)
                logging.info("Loss comparison successful")
            except AssertionError as e:
                logging.error(f"Loss comparison failed: {e}")
                raise

            try:
                torch.testing.assert_close(output_features, output_features_cp_gathered)
                logging.info("Output tensor comparison successful")
            except AssertionError as e:
                logging.error(f"Output tensor comparison failed: {e}")
                raise

            # Check gradients with and without CP.
            try:
                assert len(grads_without_cp) == len(grads_with_cp)
                logging.info(f"Comparing {len(grads_without_cp)} gradient tensors")
            except AssertionError as e:
                logging.error(f"Gradient count mismatch: {e}")
                raise

            gradient_mismatch = False
            for (n_without_cp, g_without_cp), (n_with_cp, g_with_cp) in zip(grads_without_cp, grads_with_cp):
                try:
                    torch.testing.assert_close(g_without_cp, g_with_cp)
                except AssertionError as e:
                    gradient_mismatch = True
                    logging.error(f"Gradient mismatch for {n_without_cp}: {e}")

            if gradient_mismatch:
                logging.warning("There were gradient mismatches!")
            else:
                logging.info("All gradients matched successfully!")

    finally:
        # Log final cleanup
        logging.info("Test completed, cleaning up resources")

        # Reset CUDA device
        torch.cuda.empty_cache()

        # Clean up any dangling context or process groups
        parallel_state.destroy_model_parallel()
        if dist.is_initialized():
            dist.destroy_process_group()

        # Force a small delay to ensure all cleanup is complete
        time.sleep(1)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
from pathlib import Path
from typing import Optional

import torch

from nemo.collections.asr.models.asr_model import ASRModel
from nemo.collections.audio.parts.utils.transforms import MelSpectrogram, resample
from nemo.core import Loss, Typing, typecheck
from nemo.core.neural_types import LengthsType, LossType, NeuralType, VoidType
from nemo.utils import logging
from nemo.utils.cloud import maybe_download_from_cloud
from nemo.utils.data_utils import resolve_cache_dir

from .sisnr_loss import sisnr_loss

# ASR model used for loss
# Note: Currently only this model is supported
STT_EN_CONFORMER_CTC_SMALL_v1_6_0 = 'https://api.ngc.nvidia.com/v2/models/nvidia/nemo/stt_en_conformer_ctc_small/versions/1.6.0/files/stt_en_conformer_ctc_small.nemo'


def restore_asr_model_from_cloud(location: str, refresh_cache: bool = False) -> ASRModel:
    """Restore an ASR model from the cloud.

    Args:
        location (str): The URL of the model in the cloud.
        refresh_cache (bool): Whether to force re-download of the model.

    Returns:
        nemo_asr.models.ASRModel: The restored model.
    """
    logging.debug('Restoring model from cloud location: %s', location)

    # Split into filename and base URL
    filename = location.split('/')[-1]
    url = location.replace(filename, '')

    # Get local cache dir
    cache_dir = Path.joinpath(resolve_cache_dir(), f'{filename[:-5]}')

    # If location in the cloud changes, this will force re-download
    cache_subfolder = hashlib.md5(location.encode('utf-8')).hexdigest()

    nemo_model_file_in_cache = maybe_download_from_cloud(
        url=url, filename=filename, cache_dir=cache_dir, subfolder=cache_subfolder, refresh_cache=refresh_cache
    )

    logging.debug('Model file in cache: %s', nemo_model_file_in_cache)

    # Restore model from local cache
    model = ASRModel.restore_from(nemo_model_file_in_cache)

    return model


class CombinedLoss(Loss, Typing):
    """
    Combination of three losses (signal quality/spectral+cepstral features/acoustic error)
    See https://ieeexplore.ieee.org/stamp/stamp.jsp?tp=&arnumber=1083798
    """

    def __init__(
        self,
        sample_rate: int,
        hop_length: int,
        num_mels: int,
        fft_length: int,
        sisnr_loss_weight: float,
        spectral_loss_weight: float,
        asr_loss_weight: float,
        use_asr_loss: bool = True,
        use_mel_spec: bool = True,
        conformer_model=STT_EN_CONFORMER_CTC_SMALL_v1_6_0,
        epsilon=float(5.9604644775390625e-8),
    ):

        super().__init__()
        self.sample_rate = sample_rate

        window = torch.hann_window(fft_length)
        self.register_buffer("window", window)
        epsilon = torch.tensor(epsilon)
        self.register_buffer("epsilon", epsilon, persistent=False)
        self.source_lengths = None
        self.source_value = sample_rate * sample_rate
        if use_asr_loss:
            self.asr_model = restore_asr_model_from_cloud(conformer_model)
            self.asr_model.eval()
            self.asr_model.freeze()

        self.mel_transform = MelSpectrogram(
            sample_rate=sample_rate,
            n_fft=fft_length,
            hop_length=hop_length,
            n_mels=num_mels,
            center=False,
        )
        self.mae_loss = torch.nn.L1Loss()
        self.use_mel_spec = use_mel_spec
        self.use_asr_loss = use_asr_loss

        self.sisnr_loss_weight = sisnr_loss_weight
        self.spectral_loss_weight = spectral_loss_weight
        self.asr_loss_weight = asr_loss_weight

    def spectral_loss(self, predicted_audio: torch.Tensor, primary_audio: torch.Tensor) -> torch.Tensor:
        loss = 0
        if self.use_mel_spec:
            primary_mel_spec = self.mel_transform(primary_audio)
            predicted_mel_spec = self.mel_transform(predicted_audio)
            melLoss = self.mae_loss(predicted_mel_spec, primary_mel_spec)
            loss += melLoss

            log_pred = 2 * torch.log10(torch.clamp(predicted_mel_spec, min=self.epsilon))
            log_prim = 2 * torch.log10(torch.clamp(primary_mel_spec, min=self.epsilon))
            logMelLoss = self.mae_loss(log_pred, log_prim)
            loss += logMelLoss
        return loss

    def asr_loss(self, predicted_audio: torch.Tensor, primary_audio: torch.Tensor) -> torch.Tensor:
        primary_audio_ = torch.squeeze(primary_audio, dim=1).to(next(self.parameters()).device)
        predicted_audio_ = torch.squeeze(predicted_audio, dim=1).to(next(self.parameters()).device)

        input_len = torch.full([primary_audio_.size(dim=0)], primary_audio_.size(dim=-1)).to(
            next(self.parameters()).device
        )

        asr_sample_rate = self.asr_model.cfg.sample_rate
        if self.sample_rate != asr_sample_rate:
            # Resample to 16kHz
            primary_audio_ = resample(primary_audio_, self.sample_rate, asr_sample_rate)
            predicted_audio_ = resample(predicted_audio_, self.sample_rate, asr_sample_rate)

        primary_log, _, _ = self.asr_model(input_signal=primary_audio_, input_signal_length=input_len)
        predicted_log, _, _ = self.asr_model(input_signal=predicted_audio_, input_signal_length=input_len)

        primary_prob = 10**primary_log
        predicted_prob = 10**predicted_log

        loss_fn = torch.nn.CrossEntropyLoss()
        loss = loss_fn(predicted_prob, primary_prob)
        return loss

    @property
    def input_types(self):
        """Input types definitions for CombinedLoss."""
        signal_shape = ('B', 'C', 'T')
        return {
            "estimate": NeuralType(signal_shape, VoidType()),
            "target": NeuralType(signal_shape, VoidType()),
            "input_length": NeuralType(tuple('B'), LengthsType(), optional=True),
        }

    @property
    def output_types(self):
        """Output types definitions for CombinedLoss.
        loss:
            NeuralType(None)
        """
        return {"loss": NeuralType(elements_type=LossType())}

    @typecheck()
    def forward(
        self, estimate: torch.Tensor, target: torch.Tensor, input_length: Optional[torch.Tensor] = None
    ) -> torch.Tensor:
        device = next(self.parameters()).device
        if self.source_lengths is None:
            batch = estimate.shape[0]
            self.source_lengths = torch.full((batch,), self.source_value).to(device)
        # Clip at min_len
        min_len = int(torch.min(torch.tensor([estimate.size(-1), target.size(-1)])))
        if input_length is None:
            input_length = torch.full((estimate.shape[0],), estimate.shape[1]).to(device)
        source_lengths_l = torch.where(input_length > min_len, min_len, input_length)
        primary_audio = estimate[..., :min_len]
        predicted_audio = target[..., :min_len]

        loss_total = torch.tensor([0.0]).to(device)

        # SiSNR loss
        loss_total += self.sisnr_loss_weight * sisnr_loss(primary_audio, predicted_audio, source_lengths_l)

        # Spectral Loss
        loss = self.spectral_loss(predicted_audio, primary_audio)
        loss_total += self.spectral_loss_weight * loss

        # ASR loss
        if self.use_asr_loss:
            loss_total += self.asr_loss_weight * self.asr_loss(predicted_audio, primary_audio)

        return loss_total



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from loguru import logger
from pipecat.adapters.schemas.direct_function import DirectFunction
from pipecat.adapters.schemas.tools_schema import ToolsSchema
from pipecat.processors.aggregators.openai_llm_context import OpenAILLMContext
from pipecat.services.openai.llm import OpenAILLMService


class ToolCallingMixin:
    """
    A mixin class for tool calling.
    Subclasses must implement the `setup_tool_calling` method to register all available tools
    using `self.register_direct_function()`. Then the `__init__` method of the subclass should
    call the `setup_tool_calling` method to register the tools.
    """

    def setup_tool_calling(self):
        """
        Setup the tool calling mixin by registering all available tools using self.register_direct_function().
        """
        raise NotImplementedError(
            "Subclasses must implement this method to register all available functions "
            "using self.register_direct_function()"
        )

    def register_direct_function(self, function_name: str, function: DirectFunction):
        """
        Register a direct function to be called by the LLM.

        Args:
            function_name: The name of the function to register.
            function: The direct function to register.
        """
        if not hasattr(self, "direct_functions"):
            self.direct_functions = {}
        logger.info(
            f"[{self.__class__.__name__}] Registering direct function name {function_name} to "
            f"{function.__module__ + '.' + function.__qualname__}"
        )
        self.direct_functions[function_name] = function

    @property
    def available_tools(self) -> dict[str, DirectFunction]:
        """
        Return a dictionary of available tools, where the key is the tool name and the value is the direct function.
        """
        tools = {}
        for function_name, function in self.direct_functions.items():
            tools[function_name] = function
        return tools


def register_direct_tools_to_llm(
    *,
    llm: OpenAILLMService,
    context: OpenAILLMContext,
    tool_mixins: list[ToolCallingMixin] = [],
    tools: list[DirectFunction] = [],
    cancel_on_interruption: bool = True,
) -> None:
    """
    Register direct tools to the LLM.
    Args:
        llm: The LLM service to use.
        context: The LLM context to use.
        tools: The list of tools (instances of either `DirectFunction` or `ToolCallingMixin`) to use.
    """
    all_tools = []
    for tool in tool_mixins:
        if not isinstance(tool, ToolCallingMixin):
            logger.warning(f"Tool {tool.__class__.__name__} is not a ToolCallingMixin, skipping.")
            continue
        for function_name, function in tool.available_tools.items():
            logger.info(f"Registering direct function {function_name} from {tool.__class__.__name__}")
            all_tools.append(function)

    for tool in tools:
        logger.info(f"Registering direct function: {tool.__module__ + '.' + tool.__qualname__}")
        all_tools.append(tool)

    if not all_tools:
        logger.warning("No direct tools provided.")
        return
    else:
        logger.info(f"Registering {len(all_tools)} direct tools to the LLM.")

    tools_schema = ToolsSchema(standard_tools=all_tools)
    context.set_tools(tools_schema)

    for tool in all_tools:
        llm.register_direct_function(tool, cancel_on_interruption=cancel_on_interruption)



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from nemo.utils.app_state import AppState
from nemo.utils.cast_utils import (
    CastToFloat,
    CastToFloatAll,
    avoid_bfloat16_autocast_context,
    avoid_float16_autocast_context,
    cast_all,
    cast_tensor,
    monkeypatched,
)
from nemo.utils.dtype import str_to_dtype
from nemo.utils.nemo_logging import Logger as _Logger
from nemo.utils.nemo_logging import LogMode as logging_mode

logging = _Logger()
try:
    from nemo.utils.lightning_logger_patch import add_memory_handlers_to_pl_logger

    add_memory_handlers_to_pl_logger()
except ModuleNotFoundError:
    pass

try:
    import webdataset
    from nemo.utils.data_utils import wds_url_opener

    webdataset.tariterators.url_opener = wds_url_opener
except ModuleNotFoundError:
    pass



# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn.functional as F
from torch import nn

from nemo.collections.asr.parts.submodules.jasper import jasper_activations
from nemo.core import NeuralModule
from nemo.core.neural_types import EncodedRepresentation, LossType, NeuralType


class GumbelVectorQuantizer(NeuralModule):
    def __init__(
        self,
        dim,
        num_vars,
        temp,
        groups,
        combine_groups,
        vq_dim,
        time_first,
        activation="gelu",
        weight_proj_depth=1,
        weight_proj_factor=1,
    ):
        """Vector quantization using gumbel softmax

        Args:
            dim: input dimension (channels)
            num_vars: number of quantized vectors per group
            temp: temperature for training. this should be a tuple of 3 elements: (start, stop, decay factor)
            groups: number of groups for vector quantization
            combine_groups: whether to use the vectors for all groups
            vq_dim: dimensionality of the resulting quantized vector
            time_first: if true, expect input in BxTxC format, otherwise in BxCxT
            activation: what activation to use (should be a module). this is only used if weight_proj_depth is > 1
            weight_proj_depth: number of layers (with activation in between) to project input before computing logits
            weight_proj_factor: this is used only if weight_proj_depth is > 1. scales the inner dimensionality of
                                projections by this factor
        """
        super().__init__()

        self.groups = groups
        self.combine_groups = combine_groups
        self.input_dim = dim
        self.num_vars = num_vars
        self.time_first = time_first

        assert vq_dim % groups == 0, f"dim {vq_dim} must be divisible by groups {groups} for concatenation"

        var_dim = vq_dim // groups
        num_groups = groups if not combine_groups else 1

        self.vars = nn.Parameter(torch.FloatTensor(1, num_groups * num_vars, var_dim))
        nn.init.uniform_(self.vars)

        if weight_proj_depth > 1:
            activation = jasper_activations["gelu"]

            def block(input_dim, output_dim):
                return nn.Sequential(nn.Linear(input_dim, output_dim), activation)

            inner_dim = self.input_dim * weight_proj_factor
            self.weight_proj = nn.Sequential(
                *[block(self.input_dim if i == 0 else inner_dim, inner_dim) for i in range(weight_proj_depth - 1)],
                nn.Linear(inner_dim, groups * num_vars),
            )
        else:
            self.weight_proj = nn.Linear(self.input_dim, groups * num_vars)
            nn.init.normal_(self.weight_proj.weight, mean=0, std=1)
            nn.init.zeros_(self.weight_proj.bias)

        assert len(temp) == 3, "Quantize temperature should be a tuple of 3 elements: (start, stop, decay factor)"

        self.max_temp, self.min_temp, self.temp_decay = temp
        self.curr_temp = self.max_temp
        self.codebook_indices = None

    def set_num_updates(self, num_updates):
        self.curr_temp = max(self.max_temp * self.temp_decay**num_updates, self.min_temp)

    def get_codebook_indices(self):
        if self.codebook_indices is None:
            from itertools import product

            p = [range(self.num_vars)] * self.groups
            inds = list(product(*p))
            self.codebook_indices = torch.tensor(inds, dtype=torch.long, device=self.vars.device).flatten()

            if not self.combine_groups:
                self.codebook_indices = self.codebook_indices.view(self.num_vars**self.groups, -1)
                for b in range(1, self.groups):
                    self.codebook_indices[:, b] += self.num_vars * b
                self.codebook_indices = self.codebook_indices.flatten()
        return self.codebook_indices

    def sample_from_codebook(self, b, n):
        indices = self.get_codebook_indices()
        indices = indices.view(-1, self.groups)
        cb_size = indices.size(0)
        assert n < cb_size, f"sample size {n} is greater than size of codebook {cb_size}"
        sample_idx = torch.randint(low=0, high=cb_size, size=(b * n,))
        indices = indices[sample_idx]

        z = self.vars.squeeze(0).index_select(0, indices.flatten()).view(b, n, -1)
        return z

    @property
    def input_types(self):
        """Returns definitions of module input ports."""
        if self.time_first:
            return {"x": NeuralType(('B', 'T', 'D'), EncodedRepresentation())}
        return {"x": NeuralType(('B', 'D', 'T'), EncodedRepresentation())}

    @property
    def output_types(self):
        """Returns definitions of module output ports."""
        if self.time_first:
            return {
                "x": NeuralType(('B', 'T', 'D'), EncodedRepresentation()),
                "quantize_prob_ppl": NeuralType(elements_type=LossType()),
            }
        return {
            "x": NeuralType(('B', 'D', 'T'), EncodedRepresentation()),
            "quantize_prob_ppl": NeuralType(elements_type=LossType()),
        }

    def forward(self, x, return_ids=False):

        if not self.time_first:
            x = x.transpose(1, 2)

        bsz, tsz, fsz = x.shape
        x = x.reshape(-1, fsz)
        x = self.weight_proj(x)
        x = x.view(bsz * tsz * self.groups, -1)

        _, k = x.max(-1)
        hard_x = x.new_zeros(*x.shape).scatter_(-1, k.view(-1, 1), 1.0).view(bsz * tsz, self.groups, -1)

        # Calculate quantize prob perplexity
        num_vars = self.num_vars * self.groups
        avg_probs = torch.softmax(x.view(bsz * tsz, self.groups, -1).float(), dim=-1).mean(dim=0)
        quantize_prob_ppl = torch.exp(-torch.sum(avg_probs * torch.log(avg_probs + 1e-7), dim=-1)).sum()
        quantize_prob_ppl = (num_vars - quantize_prob_ppl) / num_vars

        if self.training:
            x = F.gumbel_softmax(x.float(), tau=self.curr_temp, hard=True).type_as(x)
        else:
            x = hard_x

        x = x.view(bsz * tsz, -1)

        vars = self.vars
        if self.combine_groups:
            vars = vars.repeat(1, self.groups, 1)

        x = x.unsqueeze(-1) * vars
        x = x.view(bsz * tsz, self.groups, self.num_vars, -1)
        x = x.sum(-2)
        x = x.view(bsz, tsz, -1)

        cur_codebook_temp = self.curr_temp

        if not self.time_first:
            x = x.transpose(1, 2)  # BTC -> BCT

        if return_ids:
            hard_x_max = hard_x.argmax(-1).reshape(bsz, tsz, -1)
            # BxTxG

            # create single id from multiple group ids
            target_ids = hard_x.new_zeros(bsz, tsz).long()

            for i in range(self.groups):
                target_ids *= self.num_vars
                target_ids += hard_x_max[:, :, i]

            return x, quantize_prob_ppl, cur_codebook_temp, target_ids
        else:
            return x, quantize_prob_ppl, cur_codebook_temp



# coding=utf-8
# Copyright 2024 The Emu team, BAAI and The HuggingFace Inc. team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tokenization classes Cosmos (Visual tokens and Text tokens)."""

import base64
import logging
import os
import unicodedata
from typing import Collection, Dict, List, Optional, Set, Tuple, Union

import tiktoken
from transformers import AddedToken, PreTrainedTokenizer

logger = logging.getLogger(__name__)


VOCAB_FILES_NAMES = {
    "vocab_file": "emu3.tiktoken",
    "special_tokens_file": "cosmos_vision_tokens.txt",
}

PAT_STR = r"""(?i:'s|'t|'re|'ve|'m|'ll|'d)|[^\r\n\p{L}\p{N}]?\p{L}+|\p{N}| ?[^\s\p{L}\p{N}]+[\r\n]*|\s*[\r\n]+|\s+(?!\S)|\s+"""  # pylint: disable=line-too-long
ENDOFTEXT = "<|endoftext|>"
IMSTART = "<|im_start|>"
IMEND = "<|im_end|>"
# as the default behavior is changed to allow special tokens in
# regular texts, the surface forms of special tokens need to be
# as different as possible to minimize the impact
EXTRAS = tuple((f"<|extra_{i}|>" for i in range(205)))
# changed to use actual index to avoid misconfiguration with vocabulary expansion
SPECIAL_START_ID = 151643


def _load_tiktoken_bpe(tiktoken_bpe_file: str) -> Dict[bytes, int]:
    with open(tiktoken_bpe_file, "rb") as f:
        contents = f.read()
    return {
        base64.b64decode(token): int(rank) for token, rank in (line.split() for line in contents.splitlines() if line)
    }


class CosmosMultiModalTokenizer(PreTrainedTokenizer):
    """Emu3 tokenizer."""

    vocab_files_names = VOCAB_FILES_NAMES

    def __init__(
        self,
        vocab_file,
        special_tokens_file,
        errors="replace",
        bos_token="<|extra_203|>",
        eos_token="<|extra_204|>",
        pad_token="<|endoftext|>",
        img_token="<|image token|>",
        boi_token="<|image start|>",
        eoi_token="<|image end|>",
        eol_token="<|extra_200|>",
        eof_token="<|extra_201|>",
        **kwargs,
    ):
        super().__init__(**kwargs)

        # how to handle errors in decoding UTF-8 byte sequences
        # use ignore if you are in streaming inference
        self.errors = errors

        self.mergeable_ranks = _load_tiktoken_bpe(vocab_file)

        vision_tokens = [t.strip() for t in open(special_tokens_file).readlines() if len(t.strip()) > 0]
        SPECIAL_TOKENS = tuple(
            enumerate(
                (
                    (
                        ENDOFTEXT,
                        IMSTART,
                        IMEND,
                    )
                    + EXTRAS
                    + tuple(vision_tokens)
                ),
                start=SPECIAL_START_ID,
            )
        )
        self.special_tokens = {token: index for index, token in SPECIAL_TOKENS}
        self.special_tokens_set = set(t for _, t in SPECIAL_TOKENS)

        enc = tiktoken.Encoding(
            "Emu3",
            pat_str=PAT_STR,
            mergeable_ranks=self.mergeable_ranks,
            special_tokens=self.special_tokens,
        )

        assert (
            len(self.mergeable_ranks) + len(self.special_tokens) == enc.n_vocab
        ), f"{len(self.mergeable_ranks) + len(self.special_tokens)} != {enc.n_vocab} in encoding"

        self.decoder = {v: k for k, v in self.mergeable_ranks.items()}
        self.decoder.update({v: k for k, v in self.special_tokens.items()})

        self.tokenizer = enc

        self.eod_id = self.tokenizer.eot_token
        self.bos_token = bos_token
        self.eos_token = eos_token
        self.pad_token = pad_token
        self.img_token = img_token
        self.boi_token = boi_token
        self.eoi_token = eoi_token
        self.eol_token = eol_token
        self.eof_token = eof_token

    def __getstate__(self):
        # for pickle lovers
        state = self.__dict__.copy()
        del state["tokenizer"]
        return state

    def __setstate__(self, state):
        # tokenizer is not python native; don't pass it; rebuild it
        self.__dict__.update(state)
        enc = tiktoken.Encoding(
            "Emu3",
            pat_str=PAT_STR,
            mergeable_ranks=self.mergeable_ranks,
            special_tokens=self.special_tokens,
        )
        self.tokenizer = enc

    def __len__(self) -> int:
        return self.tokenizer.n_vocab

    def get_vocab(self) -> Dict[bytes, int]:
        """Returns the tokenizer vocab"""
        return self.mergeable_ranks

    def convert_tokens_to_ids(self, tokens: Union[bytes, str, List[Union[bytes, str]]]) -> List[int]:
        """Converts the computed tokens to their corresponding ids"""
        if isinstance(tokens, (str, bytes)):
            if tokens in self.special_tokens:
                return self.special_tokens[tokens]
            else:
                return self.mergeable_ranks.get(tokens)

        ids = []
        for token in tokens:
            if token in self.special_tokens:
                ids.append(self.special_tokens[token])
            else:
                ids.append(self.mergeable_ranks.get(token))
        return ids

    def _add_tokens(
        self,
        new_tokens: Union[List[str], List[AddedToken]],
        special_tokens: bool = False,
    ) -> int:
        if not special_tokens and new_tokens:
            raise ValueError("Adding regular tokens is not supported")

        for token in new_tokens:
            surface_form = token.content if isinstance(token, AddedToken) else token
            if surface_form not in self.special_tokens_set:
                raise ValueError("Adding unknown special tokens is not supported")

        return 0

    def save_vocabulary(self, save_directory: str, **kwargs) -> Tuple[str]:
        """
        Save only the vocabulary of the tokenizer (vocabulary).

        Returns:
            `Tuple(str)`: Paths to the files saved.
        """
        regular_file_path = os.path.join(save_directory, self.vocab_files_names["vocab_file"])
        with open(regular_file_path, 'w', encoding="utf8") as w:
            for k, v in self.mergeable_ranks.items():
                line = base64.b64encode(k).decode("utf8") + " " + str(v) + "\n"
                w.write(line)

        excluded_special_tokens = set(
            (
                ENDOFTEXT,
                IMSTART,
                IMEND,
            )
            + EXTRAS
        )
        special_file_path = os.path.join(save_directory, self.vocab_files_names["special_tokens_file"])
        with open(special_file_path, 'w', encoding="utf8") as w:
            for k in self.special_tokens:
                if k not in excluded_special_tokens:
                    print(k, file=w)

        return (regular_file_path, special_file_path)

    def tokenize(
        self,
        text: str,
        allowed_special: Union[Set, str] = "all",
        disallowed_special: Union[Collection, str] = (),
        **kwargs,
    ) -> List[Union[bytes, str]]:
        """
        Converts a string in a sequence of tokens.

        Args:
            text (`str`):
                The sequence to be encoded.
            allowed_special (`Literal["all"]` or `set`):
                The surface forms of the tokens to be encoded as special tokens in regular texts.
                Default to "all".
            disallowed_special (`Literal["all"]` or `Collection`):
                The surface forms of the tokens that should not be in regular texts and trigger errors.
                Default to an empty tuple.

            kwargs (additional keyword arguments, *optional*):
                Will be passed to the underlying model specific encode method.

        Returns:
            `List[bytes|str]`: The list of tokens.
        """
        tokens = []
        text = unicodedata.normalize("NFC", text)

        # this implementation takes a detour: text -> token id -> token surface forms
        for t in self.tokenizer.encode(text, allowed_special=allowed_special, disallowed_special=disallowed_special):
            tokens.append(self.decoder[t])

        return tokens

    def convert_tokens_to_string(self, tokens: List[Union[bytes, str]]) -> str:
        """
        Converts a sequence of tokens in a single string.
        """
        text = ""
        temp = b""
        for t in tokens:
            if isinstance(t, str):
                if temp:
                    text += temp.decode("utf-8", errors=self.errors)
                    temp = b""
                text += t
            elif isinstance(t, bytes):
                temp += t
            else:
                raise TypeError("token should only be of type types or str")
        if temp:
            text += temp.decode("utf-8", errors=self.errors)
        return text

    @property
    def vocab_size(self):
        """Returns the vocab size"""
        return self.tokenizer.n_vocab

    def _convert_id_to_token(self, index: int) -> Union[bytes, str]:
        """Converts an id to a token, special tokens included"""
        if index in self.decoder:
            return self.decoder[index]
        raise ValueError("unknown ids")

    def _convert_token_to_id(self, token: Union[bytes, str]) -> int:
        """Converts a token to an id using the vocab, special tokens included"""
        if token in self.special_tokens:
            return self.special_tokens[token]
        if token in self.mergeable_ranks:
            return self.mergeable_ranks[token]
        raise ValueError("unknown token")

    def _decode(
        self,
        token_ids: Union[int, List[int]],
        skip_special_tokens: bool = False,
        errors: Optional[str] = None,
        **kwargs,
    ) -> str:
        if isinstance(token_ids, int):
            token_ids = [token_ids]

        if skip_special_tokens:
            token_ids = [i for i in token_ids if i < self.eod_id]

        return self.tokenizer.decode(token_ids, errors=errors or self.errors)



# Copyright (c) 2023, NVIDIA CORPORATION & AFFILIATES.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Tuple

import librosa
import numpy as np
import torch

from nemo.collections.asr.models import EncDecClassificationModel
from nemo.collections.tts.parts.utils.tts_dataset_utils import normalize_volume
from nemo.utils import logging


class AudioTrimmer(ABC):
    """Interface for silence trimming implementations
    """

    @abstractmethod
    def trim_audio(self, audio: np.array, sample_rate: int, audio_id: str) -> Tuple[np.array, int, int]:
        """Trim starting and trailing silence from the input audio.
           Args:
               audio: Numpy array containing audio samples. Float [-1.0, 1.0] format.
               sample_rate: Sample rate of input audio.
               audio_id: String identifier (eg. file name) used for logging.

           Returns numpy array with trimmed audio, and integer sample indices representing the start and end
           of speech within the original audio array.
        """
        raise NotImplementedError


class EnergyAudioTrimmer(AudioTrimmer):
    def __init__(
        self,
        db_threshold: int = 50,
        ref_amplitude: float = 1.0,
        speech_frame_threshold: int = 1,
        trim_win_length: int = 2048,
        trim_hop_length: int = 512,
        pad_seconds: float = 0.1,
        volume_norm: bool = True,
    ) -> None:
        """Energy/power based silence trimming using Librosa backend.
           Args:
               db_threshold: Audio frames at least db_threshold decibels below ref_amplitude will be
                 considered silence.
               ref_amplitude: Amplitude threshold for classifying speech versus silence.
               speech_frame_threshold: Start and end of speech will be detected where there are at least
                 speech_frame_threshold consecutive audio frames classified as speech. Setting this value higher
                 is more robust to false-positives (silence detected as speech), but setting it too high may result
                 in very short speech segments being cut out from the audio.
               trim_win_length: Length of audio frames to use when doing speech detection. This does not need to match
                 the win_length used any other part of the code or model.
               trim_hop_length: Stride of audio frames to use when doing speech detection. This does not need to match
                 the hop_length used any other part of the code or model.
               pad_seconds: Audio duration in seconds to keep before and after each speech segment.
                 Set this to at least 0.1 to avoid cutting off any speech audio, with larger values
                 being safer but increasing the average silence duration left afterwards.
               volume_norm: Whether to normalize the volume of audio before doing speech detection.
        """
        assert db_threshold >= 0
        assert ref_amplitude >= 0
        assert speech_frame_threshold > 0
        assert trim_win_length > 0
        assert trim_hop_length > 0

        self.db_threshold = db_threshold
        self.ref_amplitude = ref_amplitude
        self.speech_frame_threshold = speech_frame_threshold
        self.trim_win_length = trim_win_length
        self.trim_hop_length = trim_hop_length
        self.pad_seconds = pad_seconds
        self.volume_norm = volume_norm

    def trim_audio(self, audio: np.array, sample_rate: int, audio_id: str = "") -> Tuple[np.array, int, int]:
        if self.volume_norm:
            # Normalize volume so we have a fixed scale relative to the reference amplitude
            audio = normalize_volume(audio=audio, volume_level=1.0)

        speech_frames = librosa.effects._signal_to_frame_nonsilent(
            audio,
            ref=self.ref_amplitude,
            frame_length=self.trim_win_length,
            hop_length=self.trim_hop_length,
            top_db=self.db_threshold,
        )

        start_frame, end_frame = get_start_and_end_of_speech_frames(
            is_speech=speech_frames, speech_frame_threshold=self.speech_frame_threshold, audio_id=audio_id,
        )
        if not start_frame and not end_frame:
            return np.array([]), 0, 0

        start_sample = librosa.core.frames_to_samples(start_frame, hop_length=self.trim_hop_length)
        end_sample = librosa.core.frames_to_samples(end_frame, hop_length=self.trim_hop_length)

        start_sample, end_sample = pad_sample_indices(
            start_sample=start_sample,
            end_sample=end_sample,
            max_sample=audio.shape[0],
            sample_rate=sample_rate,
            pad_seconds=self.pad_seconds,
        )

        trimmed_audio = audio[start_sample:end_sample]

        return trimmed_audio, start_sample, end_sample


class VadAudioTrimmer(AudioTrimmer):
    def __init__(
        self,
        model_name: str = "vad_multilingual_marblenet",
        vad_sample_rate: int = 16000,
        vad_threshold: float = 0.5,
        device: str = "cpu",
        speech_frame_threshold: int = 1,
        trim_win_length: int = 4096,
        trim_hop_length: int = 1024,
        pad_seconds: float = 0.1,
        volume_norm: bool = True,
    ) -> None:
        """Voice activity detection (VAD) based silence trimming.

           Args:
               model_name: NeMo VAD model to load. Valid configurations can be found with
                 EncDecClassificationModel.list_available_models()
               vad_sample_rate: Sample rate used for pretrained VAD model.
               vad_threshold: Softmax probability [0, 1] of VAD output, above which audio frames will be classified
                 as speech.
               device: Device "cpu" or "cuda" to use for running the VAD model.
               trim_win_length: Length of audio frames to use when doing speech detection. This does not need to match
                 the win_length used any other part of the code or model.
               trim_hop_length: Stride of audio frames to use when doing speech detection. This does not need to match
                 the hop_length used any other part of the code or model.
               pad_seconds: Audio duration in seconds to keep before and after each speech segment.
                 Set this to at least 0.1 to avoid cutting off any speech audio, with larger values
                 being safer but increasing the average silence duration left afterwards.
               volume_norm: Whether to normalize the volume of audio before doing speech detection.
        """
        assert vad_sample_rate > 0
        assert vad_threshold >= 0
        assert speech_frame_threshold > 0
        assert trim_win_length > 0
        assert trim_hop_length > 0

        self.device = device
        self.vad_model = EncDecClassificationModel.from_pretrained(model_name=model_name).eval().to(self.device)
        self.vad_sample_rate = vad_sample_rate
        self.vad_threshold = vad_threshold

        self.speech_frame_threshold = speech_frame_threshold
        self.trim_win_length = trim_win_length
        self.trim_hop_length = trim_hop_length
        # Window shift neeeded in order to center frames
        self.trim_shift = self.trim_win_length // 2

        self.pad_seconds = pad_seconds
        self.volume_norm = volume_norm

    def _detect_speech(self, audio: np.array) -> np.array:
        if audio.shape[0] < self.trim_win_length:
            return np.array([])

        # [num_frames, win_length]
        audio_frames = librosa.util.frame(
            audio, frame_length=self.trim_win_length, hop_length=self.trim_hop_length
        ).transpose()
        audio_frame_lengths = audio_frames.shape[0] * [self.trim_win_length]

        # [num_frames, win_length]
        audio_signal = torch.tensor(audio_frames, dtype=torch.float32, device=self.device)
        # [1]
        audio_signal_len = torch.tensor(audio_frame_lengths, dtype=torch.int32, device=self.device)
        # VAD outputs 2 values for each audio frame with logits indicating the likelihood that
        # each frame is non-speech or speech, respectively.
        # [num_frames, 2]
        log_probs = self.vad_model(input_signal=audio_signal, input_signal_length=audio_signal_len)
        probs = torch.softmax(log_probs, dim=-1)
        probs = probs.detach().cpu().numpy()
        # [num_frames]
        speech_probs = probs[:, 1]
        speech_frames = speech_probs >= self.vad_threshold

        return speech_frames

    def _scale_sample_indices(self, start_sample: int, end_sample: int, sample_rate: int) -> Tuple[int, int]:
        sample_rate_ratio = sample_rate / self.vad_sample_rate
        start_sample = int(sample_rate_ratio * start_sample)
        end_sample = int(sample_rate_ratio * end_sample)
        return start_sample, end_sample

    def trim_audio(self, audio: np.array, sample_rate: int, audio_id: str = "") -> Tuple[np.array, int, int]:
        if sample_rate == self.vad_sample_rate:
            vad_audio = audio
        else:
            # Resample audio to match sample rate of VAD model
            vad_audio = librosa.resample(audio, orig_sr=sample_rate, target_sr=self.vad_sample_rate)

        if self.volume_norm:
            # Normalize volume so we have a fixed scale relative to the reference amplitude
            vad_audio = normalize_volume(audio=vad_audio, volume_level=1.0)

        speech_frames = self._detect_speech(audio=vad_audio)

        start_frame, end_frame = get_start_and_end_of_speech_frames(
            is_speech=speech_frames, speech_frame_threshold=self.speech_frame_threshold, audio_id=audio_id,
        )
        if not start_frame and not end_frame:
            return np.array([]), 0, 0

        if start_frame == 0:
            start_sample = 0
        else:
            start_sample = librosa.core.frames_to_samples(start_frame, hop_length=self.trim_hop_length)
            start_sample += self.trim_shift

        # Avoid trimming off the end because VAD model is not trained to classify partial end frames.
        if end_frame == speech_frames.shape[0]:
            end_sample = vad_audio.shape[0]
        else:
            end_sample = librosa.core.frames_to_samples(end_frame, hop_length=self.trim_hop_length)
            end_sample += self.trim_shift

        if sample_rate != self.vad_sample_rate:
            # Convert sample indices back to input sample rate
            start_sample, end_sample = self._scale_sample_indices(
                start_sample=start_sample, end_sample=end_sample, sample_rate=sample_rate
            )

        start_sample, end_sample = pad_sample_indices(
            start_sample=start_sample,
            end_sample=end_sample,
            max_sample=audio.shape[0],
            sample_rate=sample_rate,
            pad_seconds=self.pad_seconds,
        )

        trimmed_audio = audio[start_sample:end_sample]

        return trimmed_audio, start_sample, end_sample


def get_start_and_end_of_speech_frames(
    is_speech: np.array, speech_frame_threshold: int, audio_id: str = ""
) -> Tuple[int, int]:
    """Finds the speech frames corresponding to the start and end of speech for an utterance.
       Args:
           is_speech: [num_frames] boolean array with true entries labeling speech frames.
           speech_frame_threshold: The number of consecutive speech frames required to classify the speech boundaries.
           audio_id: String identifier (eg. file name) used for logging.

       Returns integers representing the frame indices of the start (inclusive) and end (exclusive) of speech.
    """
    num_frames = is_speech.shape[0]

    # Iterate forwards over the utterance until we find the first speech_frame_threshold consecutive speech frames.
    start_frame = None
    for i in range(0, num_frames - speech_frame_threshold + 1):
        high_i = i + speech_frame_threshold
        if all(is_speech[i:high_i]):
            start_frame = i
            break

    # Iterate backwards over the utterance until we find the last speech_frame_threshold consecutive speech frames.
    end_frame = None
    for i in range(num_frames, speech_frame_threshold - 1, -1):
        low_i = i - speech_frame_threshold
        if all(is_speech[low_i:i]):
            end_frame = i
            break

    if start_frame is None or end_frame is None:
        # Algorithm is symmetric, so if the start is not found then the end should also not be found.
        logging.warning(f"Could not find start or end of speech for '{audio_id}'")
        return 0, 0

    return start_frame, end_frame


def pad_sample_indices(
    start_sample: int, end_sample: int, max_sample: int, sample_rate: int, pad_seconds: float
) -> Tuple[int, int]:
    """Shift the input sample indices by pad_seconds in front and back within [0, max_sample]
       Args:
           start_sample: Start sample index
           end_sample: End sample index
           max_sample: Maximum sample index
           sample_rate: Sample rate of audio
           pad_seconds: Amount to pad/shift the indices by.

       Returns the sample indices after padding by the input amount.
    """
    pad_samples = int(pad_seconds * sample_rate)
    start_sample = start_sample - pad_samples
    end_sample = end_sample + pad_samples

    start_sample = max(0, start_sample)
    end_sample = min(max_sample, end_sample)

    return start_sample, end_sample



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib


# @akoumparouli: use a context manager that saves/restores gbs/mbs when using
# reconfigure_num_microbatches_calculator to avoid interference between tests.
@contextlib.contextmanager
def reconfigure_num_microbatches_calculator_manager(*args, **kwargs):
    import megatron.core.num_microbatches_calculator as mb_calc

    # Store current mbs, gbs values
    if not mb_calc._GLOBAL_NUM_MICROBATCHES_CALCULATOR is None:
        _mbs = mb_calc.get_micro_batch_size()
        _gbs = mb_calc.get_current_global_batch_size()

        # use user's settings
        mb_calc.reconfigure_num_microbatches_calculator(*args, **kwargs)
    else:
        _mbs, _gbs = 1, 1

    try:
        # run user's code
        yield
        # @akoumparouli: no catch
    finally:
        # restore old mbs, gbs
        if not mb_calc._GLOBAL_NUM_MICROBATCHES_CALCULATOR is None:
            mb_calc.reconfigure_num_microbatches_calculator(0, None, _gbs, _mbs, data_parallel_size=1)



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import fiddle as fdl

from nemo.lightning.io import drop_unexpected_params


class TestDropUnexpectedParams:

    def setup_method(self):
        """
        Setup common test resources.
        """

        class MockClassOld:
            def __init__(self, x, y, deprecated):
                pass

        class MockClassNew:
            def __init__(self, x, y):
                pass

        class OuterClass:
            def __init__(self, z, t):
                pass

        self.MockClassOld = MockClassOld
        self.MockClassNew = MockClassNew
        self.OuterClass = OuterClass

    def test_valid_config_stays_same(self):
        """
        Test that a valid config remains unchanged.
        """

        config = fdl.Config(self.MockClassNew, x=1, y=2)
        updated = drop_unexpected_params(config)

        assert not updated, "Expected the config to remain unchanged."
        assert config.x == 1
        assert config.y == 2

    def test_config_updates(self):
        """
        Test that a config with unexpected parameters gets updated.
        """
        config = fdl.Config(self.MockClassOld, x=1, y=2, deprecated=3)

        # Simulate deprecation issue by overriding target class
        config.__dict__['__fn_or_cls__'] = self.MockClassNew

        updated = drop_unexpected_params(config)
        assert updated, "Expected the config to be updated."
        assert config.x == 1
        assert config.y == 2
        assert not hasattr(config, "deprecated"), "Expected 'deprecated' to be removed from the config."

    def test_nested_config_updates(self):
        """
        Test that a nested config with unexpected parameters gets updated.
        """
        config = fdl.Config(self.OuterClass, z=4, t=fdl.Config(self.MockClassOld, x=1, y=2, deprecated=3))

        # Simulate deprecation issue by overriding target class
        config.t.__dict__["__fn_or_cls__"] = self.MockClassNew

        updated = drop_unexpected_params(config)
        assert updated, "Expected the nested config to be updated."
        assert config.z == 4
        assert config.t.x == 1
        assert config.t.y == 2
        assert not hasattr(config.t, "deprecated"), "Expected 'deprecated' to be removed from the inner config."



# Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# Copyright (c) 2024 Arc Institute. All rights reserved.
# Copyright (c) 2024 Michael Poli. All rights reserved.
# Copyright (c) 2024 Stanford University. All rights reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from nemo.collections import llm
from nemo.collections.llm.gpt.model.hyena import (
    HuggingFaceSavannaHyenaImporter,
    Hyena1bConfig,
    Hyena7bARCLongContextConfig,
    Hyena7bConfig,
    Hyena40bARCLongContextConfig,
    Hyena40bConfig,
    HyenaConfig,
    HyenaNV1bConfig,
    HyenaNV7bConfig,
    HyenaNV40bConfig,
    HyenaNVTestConfig,
    HyenaTestConfig,
)


def test_hyena_base_config():
    config = HyenaConfig()
    assert config.num_layers == 2
    assert config.hidden_size == 1024
    assert config.num_attention_heads == 8
    assert config.seq_length == 2048
    assert config.position_embedding_type == "rope"
    assert config.make_vocab_size_divisible_by == 128
    assert config.num_groups_hyena == None
    assert config.num_groups_hyena_medium == None
    assert config.num_groups_hyena_short == None
    assert config.hybrid_attention_ratio == 0.0
    assert config.hybrid_mlp_ratio == 0.0
    assert config.gated_linear_unit == True


def test_hyena_7b_config():
    config = Hyena7bConfig()
    assert config.num_layers == 32
    assert config.hidden_size == 4096
    assert config.num_attention_heads == 32
    assert config.seq_length == 8192
    assert config.num_groups_hyena == 4096
    assert config.num_groups_hyena_medium == 256
    assert config.num_groups_hyena_short == 256
    assert config.hybrid_override_pattern == "SDH*SDHSDH*SDHSDH*SDHSDH*SDHSDH*"


def test_hyena_nv_7b_config():
    config = HyenaNV7bConfig()
    assert config.num_layers == 32
    assert config.hidden_size == 4096
    assert config.num_attention_heads == 32
    assert config.seq_length == 8192
    assert config.tokenizer_library == "byte-level"
    assert config.mapping_type == "base"
    assert config.use_short_conv_bias is True


def test_hyena_1b_config():
    config = Hyena1bConfig()
    assert config.num_layers == 25
    assert config.hidden_size == 1920
    assert config.num_attention_heads == 15
    assert config.seq_length == 8192
    assert config.num_groups_hyena == 1920
    assert config.num_groups_hyena_medium == 128
    assert config.num_groups_hyena_short == 128


def test_hyena_nv_1b_config():
    config = HyenaNV1bConfig()
    assert config.num_layers == 25
    assert config.hidden_size == 1920
    assert config.ffn_hidden_size == 5120
    assert config.num_attention_heads == 15
    assert config.seq_length == 8192
    assert config.tokenizer_library == "byte-level"


def test_hyena_40b_config():
    config = Hyena40bConfig()
    assert config.num_layers == 50
    assert config.hidden_size == 8192
    assert config.num_attention_heads == 64
    assert config.seq_length == 8192
    assert config.num_groups_hyena == 8192
    assert config.num_groups_hyena_medium == 512
    assert config.num_groups_hyena_short == 512


def test_hyena_nv_40b_config():
    config = HyenaNV40bConfig()
    assert config.num_layers == 50
    assert config.hidden_size == 8192
    assert config.ffn_hidden_size == 22528
    assert config.num_attention_heads == 64
    assert config.seq_length == 8192
    assert config.tokenizer_library == "byte-level"


def test_hyena_7b_arc_long_context_config():
    config = Hyena7bARCLongContextConfig()
    assert config.num_layers == 32
    assert config.hidden_size == 4096
    assert config.ffn_hidden_size == 11264
    assert config.num_attention_heads == 32
    assert config.seq_length == 1048576  # ~1M or 2**20
    assert config.tokenizer_library == "byte-level"


def test_hyena_40b_arc_long_context_config():
    config = Hyena40bARCLongContextConfig()
    assert config.num_layers == 50
    assert config.hidden_size == 8192
    assert config.ffn_hidden_size == 22528
    assert config.num_attention_heads == 64
    assert config.seq_length == 1048576  # ~1M or 2**20
    assert config.tokenizer_library == "byte-level"


def test_hyena_test_config():
    config = HyenaTestConfig()
    assert config.num_layers == 4
    assert config.hidden_size == 4096
    assert config.num_attention_heads == 32
    assert config.seq_length == 8192
    assert config.hybrid_override_pattern == "SDH*"


def test_hyena_nv_test_config():
    config = HyenaNVTestConfig()
    assert config.num_layers == 4
    assert config.hidden_size == 4096
    assert config.ffn_hidden_size == 11008
    assert config.num_attention_heads == 32
    assert config.seq_length == 8192
    assert config.tokenizer_library == "byte-level"


@pytest.mark.pleasefixme
def test_convert_hyena():
    from huggingface_hub.utils import RepositoryNotFoundError

    evo2_config = llm.Hyena1bConfig()
    model_ckpt = "dummy_model_deosnt_exist"
    exporter = HuggingFaceSavannaHyenaImporter(model_ckpt, model_config=evo2_config)

    with pytest.raises(RepositoryNotFoundError):
        exporter.apply("dummy_output_deosnt_exist")



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from unittest.mock import MagicMock

import pytest
from megatron.core import parallel_state
from torch import nn

from nemo import lightning as nl
from nemo.lightning import megatron_parallel as mp


class TestMegatronParallel:
    """Unit tests for the MegatronParallel class."""

    @pytest.fixture
    def mock_pipeline(self, mocker):
        """Fixture to create a mock pipeline."""

        class DummyModule(nn.Module):
            def __init__(self, dummy_arg=None):
                self.dummy_arg = dummy_arg
                super().__init__()

            def forward(self, x):
                return x

        return DummyModule()

    @pytest.fixture
    def mock_precision_plugin(self, mocker):
        """Fixture to create a mock precision plugin."""
        return nl.MegatronMixedPrecision(precision="bf16-mixed")

    @pytest.fixture
    def mock_callbacks(self, mocker):
        """Fixture to create a mock callback connector."""
        return mocker.MagicMock(spec=mp.CallbackConnector)

    @pytest.fixture
    def mock_data_step(self, mocker):
        """Fixture to create a mock data step function."""
        return mocker.MagicMock()

    @pytest.fixture
    def mock_forward_step(self, mocker):
        """Fixture to create a mock forward step function."""
        return mocker.MagicMock()

    @pytest.fixture
    def mock_loss_reduction(self, mocker):
        """Fixture to create a mock loss reduction function."""
        return mocker.MagicMock()

    def test_init_with_defaults(self, mocker, mock_pipeline):
        """Test __init__ with default parameters."""
        mocker.patch('megatron.core.parallel_state.get_pipeline_model_parallel_world_size', return_value=1)
        mocker.patch('megatron.core.parallel_state.model_parallel_is_initialized', return_value=False)

        megatron_parallel = mp.MegatronParallel(pipeline=mock_pipeline, cpu=True)

        assert megatron_parallel.pipeline == mock_pipeline
        assert megatron_parallel.precision_plugin is None
        assert isinstance(megatron_parallel.callbacks, mp.CallbackConnector)
        assert megatron_parallel.data_step == mp.default_data_step
        assert megatron_parallel.forward_step == mp.default_forward_step
        assert megatron_parallel.loss_reduction is None

    def test_init_with_custom_parameters(
        self,
        mocker,
        mock_pipeline,
        mock_precision_plugin,
        mock_callbacks,
        mock_data_step,
        mock_forward_step,
        mock_loss_reduction,
    ):
        """Test __init__ with custom parameters."""
        mocker.patch('megatron.core.parallel_state.get_pipeline_model_parallel_world_size', return_value=1)
        mocker.patch('megatron.core.parallel_state.model_parallel_is_initialized', return_value=False)

        megatron_parallel = mp.MegatronParallel(
            pipeline=mock_pipeline,
            precision_plugin=mock_precision_plugin,
            callbacks=mock_callbacks,
            data_step=mock_data_step,
            forward_step=mock_forward_step,
            loss_reduction=mock_loss_reduction,
            cpu=True,
        )

        assert megatron_parallel.pipeline == mock_pipeline
        assert megatron_parallel.precision_plugin == mock_precision_plugin
        assert megatron_parallel.callbacks == mock_callbacks
        assert megatron_parallel.data_step == mock_data_step
        assert megatron_parallel.forward_step == mock_forward_step
        assert megatron_parallel.loss_reduction == mock_loss_reduction


class TestCallbackConnector:
    def test_add_callbacks(self) -> None:
        callback_connector = mp.CallbackConnector()
        callback = TestCallback()
        callback_connector.add(callback)

        assert callback in callback_connector.callbacks["on_megatron_step_start"]
        assert callback in callback_connector.callbacks["on_megatron_microbatch_start"]

    def test_event(self) -> None:
        callback_connector = mp.CallbackConnector()
        callback = TestCallback()
        callback_connector.add(callback)

        # Replace mocker.spy with manual mocking
        callback.on_megatron_step_start = MagicMock()
        callback.on_megatron_microbatch_start = MagicMock()

        callback_connector.event("on_megatron_step_start")
        callback_connector.event("on_megatron_microbatch_start")

        assert callback.on_megatron_step_start.call_count == 1
        assert callback.on_megatron_microbatch_start.call_count == 1

    def test_add_connector(self) -> None:
        callback_connector1 = mp.CallbackConnector()
        callback_connector2 = mp.CallbackConnector()
        callback1 = TestCallback()
        callback2 = TestCallback()

        callback_connector1.add(callback1)
        callback_connector2.add(callback2)

        callback_connector1 += callback_connector2

        assert callback1 in callback_connector1.callbacks["on_megatron_step_start"]
        assert callback2 in callback_connector1.callbacks["on_megatron_step_start"]

    def test_contains(self):
        callback_connector = mp.CallbackConnector()
        callback = TestCallback()
        callback_connector.add(callback)

        assert callback in callback_connector

    def test_add_count_callback(self):
        """Test adding a CountCallback to the CallbackConnector."""
        connector = mp.CallbackConnector()
        count_callback = CountCallback()
        connector.add(count_callback)

        # Check if the CountCallback has been added correctly
        assert count_callback in connector, "CountCallback should be in the CallbackConnector"

    def test_event_trigger_with_count_callback(self):
        """Test if the event triggers the method in CountCallback."""
        connector = mp.CallbackConnector()
        count_callback = CountCallback()
        connector.add(count_callback)

        # Simulate an event that CountCallback listens to
        connector.event('on_megatron_step_start')

        # Check if the CountCallback's method was called
        assert (
            count_callback.counts["on_megatron_step_start"] == 1
        ), "CountCallback's method should have been triggered once"


class TestCallback:
    def on_megatron_step_start(self):
        pass

    def on_megatron_microbatch_start(self):
        pass


class CountCallback:
    def __init__(self) -> None:
        self.counts = defaultdict(int)

    def on_megatron_step_start(self, *args, **kwargs) -> None:
        # assert len(kwargs) == 12
        self.counts["on_megatron_step_start"] += 1

    def on_megatron_microbatch_start(self, *args, **kwargs) -> None:
        # assert len(kwargs) == 14
        self.counts["on_megatron_microbatch_start"] += 1

    def on_megatron_microbatch_callback(self, *args, **kwargs) -> None:
        self.counts["on_megatron_microbatches_callback"] += 1

    def on_megatron_microbatch_end(self, *args, **kwargs) -> None:
        self.counts["on_megatron_microbatches_end"] += 1

    def on_megatron_reduce_microbatches_start(self, *args, **kwargs) -> None:
        self.counts["on_megatron_reduce_microbatches_start"] += 1

    def on_megatron_reduce_microbatches_end(self, *args, **kwargs) -> None:
        self.counts["on_megatron_reduce_microbatches_end"] += 1

    def on_megatron_log_step_end(self, *args, **kwargs) -> None:
        self.counts["on_megatron_log_step_end"] += 1

    def on_megatron_step_end(self, *args, **kwargs) -> None:
        self.counts["on_megatron_step_end"] += 1



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest
import torch
from lhotse import CutSet, MonoCut, SupervisionSegment
from lhotse.testing.dummies import dummy_recording
from omegaconf import OmegaConf

from nemo.collections.common.data.lhotse.dataloader import get_lhotse_dataloader_from_config
from nemo.collections.common.data.lhotse.text_adapters import SourceTargetTextExample
from nemo.collections.common.tokenizers import SentencePieceTokenizer
from nemo.collections.common.tokenizers.sentencepiece_tokenizer import create_spt_model
from nemo.collections.multimodal.speech_llm.data.lhotse_dataset import LhotseAudioQuestionAnswerDataset
from nemo.collections.multimodal.speech_llm.parts.utils.data_utils import PromptFormatterTextProcessing


class Identity(torch.utils.data.Dataset):
    def __getitem__(self, cuts):
        return cuts


@pytest.fixture
def tokenizer(capsys, tmp_path_factory):
    TOKENIZER_TRAIN_TEXT = """
    Example system message.
    Example user message.
    Example assistant message.
    TEST
    [INST]
    [/INST]
    <s>
    </s>
    <<SYS>>
    <</SYS>>
    User: Assistant:
    user model
    Instruct Output
    \n\n
    <start_of_turn> <end_of_turn>
    <|
    |>
    <|en|> <|de|> <|fr|> <|es|> <|transcribe|> <|translate|> <|pnc|> <|nopnc|> <|startoftranscript|> <|endoftext|>
    Feel free to add new tokens for your own tests!?
    But know that if you do so, you may need to update the token IDs in the existing tests!
    So, it might be a good idea to create a new tokenizer instead when adding new prompt formats.
    """
    tmpdir = tmp_path_factory.mktemp("bpe_tokenizer")
    text_path = tmpdir / "text.txt"
    text_path.write_text(TOKENIZER_TRAIN_TEXT)
    with capsys.disabled():
        create_spt_model(
            str(text_path),
            vocab_size=512,
            sample_size=-1,
            do_lower_case=False,
            output_dir=str(tmpdir),
            remove_extra_whitespaces=True,
        )
    return SentencePieceTokenizer(str(tmpdir / "tokenizer.model"))


"""
TEST FOR AUDIO DATALOADING WITH EMMETT
"""


@pytest.fixture
def cuts():
    return CutSet(
        [
            MonoCut(
                id="ex0",
                start=0,
                duration=5.0,
                channel=0,
                supervisions=[
                    SupervisionSegment(
                        id="ex0",
                        recording_id="dummy-recording-0000",
                        start=0,
                        duration=5.0,
                        text="some transcription",
                        language="en",
                    )
                ],
                recording=dummy_recording(0, duration=5.0, with_data=True),
                custom={
                    "context": "<en>",
                    "answer": "some desired answer",
                },
            ),
        ]
    )


@pytest.fixture
def cuts_path(tmp_path_factory, cuts):
    tmp_path = tmp_path_factory.mktemp("data")
    p = tmp_path / "cuts.jsonl.gz"
    pa = tmp_path / "audio"
    cuts.save_audios(pa).to_file(p)
    return p


def test_audio_example_with_prompt_emmett_t5(cuts_path, tokenizer):
    config = OmegaConf.create(
        {
            "input_cfg": [
                {
                    "type": "lhotse",
                    "cuts_path": cuts_path,
                },
            ],
            "prompt_format": "t5nmt",
            "force_finite": True,
            "shuffle": True,
            "num_workers": 0,
            "batch_size": 1,
            "seed": 0,
            "shard_seed": 0,
        }
    )

    # First test that sampling is correct and tokenizer + prompt formatter is applied there

    dl = get_lhotse_dataloader_from_config(
        config=config, global_rank=0, world_size=1, dataset=Identity(), tokenizer=tokenizer
    )
    batches = [batch for batch in dl]
    assert len(batches) == 1

    b = batches[0]
    assert isinstance(b, CutSet)
    assert len(b) == 1
    ex = b[0]
    assert isinstance(ex, MonoCut)

    assert ex.has_custom("context_ids")
    assert torch.is_tensor(ex.context_ids)
    assert tokenizer.ids_to_text(ex.context_ids) == "<en>"

    assert ex.has_custom("answer_ids")
    assert torch.is_tensor(ex.answer_ids)
    assert tokenizer.ids_to_text(ex.answer_ids) == "some transcription"

    assert ex.has_custom("input_ids")
    assert torch.is_tensor(ex.input_ids)
    assert tokenizer.ids_to_text(ex.input_ids) == "<en> some transcription"

    # Test that speechlm dataset processes the example correctly

    text_processor = PromptFormatterTextProcessing(tokenizer=tokenizer, prompt_format="t5nmt")
    dataset = LhotseAudioQuestionAnswerDataset(
        text_processor=text_processor,
        default_context="<en>",
        tokens_to_generate=0,
        pad_to_max_length=False,
        max_seq_length=64,
    )

    batch = dataset[batches[0]]
    assert tokenizer.ids_to_text(batch["tokens"][0]) == "<en> some transcriptio"
    assert tokenizer.ids_to_text(batch["labels"][0]) == "en> some transcription"
    assert tokenizer.ids_to_text(batch["contexts"][0]) == "<en>"
    assert tokenizer.ids_to_text(batch["answers"][0]) == "some transcription"


"""
TEST FOR TEXT DATALOADING WITH EMMETT
"""


@pytest.fixture
def nmt_paths(tmp_path_factory):
    tmp_path = tmp_path_factory.mktemp("nmtdata")
    src = tmp_path / "src.txt"
    tgt = tmp_path / "tgt.txt"
    q = tmp_path / "q.txt"
    src.write_text("fake german")
    tgt.write_text("real english")
    q.write_text("<en>")
    return src, tgt, q


def test_text_example_with_prompt_emmett_t5(nmt_paths, tokenizer):
    src, tgt, q = nmt_paths
    config = OmegaConf.create(
        {
            "input_cfg": [
                {
                    "type": "txt_pair",
                    "source_paths": src,
                    "target_paths": tgt,
                    "source_language": "de",
                    "target_language": "en",
                    "questions_path": q,
                    "questions_language": "en",
                },
            ],
            "prompt_format": "t5nmt",
            "force_finite": True,
            "shuffle": True,
            "num_workers": 0,
            "batch_size": 1,
            "seed": 0,
            "shard_seed": 0,
        }
    )

    # First test that sampling is correct and tokenizer + prompt formatter is applied there

    dl = get_lhotse_dataloader_from_config(
        config=config, global_rank=0, world_size=1, dataset=Identity(), tokenizer=tokenizer
    )
    batches = [batch for batch in dl]
    assert len(batches) == 1

    b = batches[0]
    assert isinstance(b, CutSet)
    assert len(b) == 1
    ex = b[0]
    assert isinstance(ex, SourceTargetTextExample)

    assert torch.is_tensor(ex.context_ids)
    assert tokenizer.ids_to_text(ex.context_ids) == "<en> fake german"

    assert torch.is_tensor(ex.answer_ids)
    assert tokenizer.ids_to_text(ex.answer_ids) == "real english"

    assert torch.is_tensor(ex.input_ids)
    assert tokenizer.ids_to_text(ex.input_ids) == "<en> fake german real english"

    # Test that speechlm dataset processes the example correctly

    text_processor = PromptFormatterTextProcessing(tokenizer=tokenizer, prompt_format="t5nmt")
    dataset = LhotseAudioQuestionAnswerDataset(
        text_processor=text_processor,
        default_context="<en>",
        tokens_to_generate=0,
        pad_to_max_length=False,
        max_seq_length=64,
    )

    batch = dataset[batches[0]]

    assert tokenizer.ids_to_text(batch["text_input_ids"][0]) == "<en> fake german real english"
    assert tokenizer.ids_to_text(batch["text_context_ids"][0]) == "<en> fake german"
    assert tokenizer.ids_to_text(batch["text_answer_ids"][0]) == "real english"



# SPDX-FileCopyrightText: Copyright (c) 2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for the inference libraries."""

import os
import re
from glob import glob

import mediapy as media
import numpy as np
import torch

from nemo.collections.common.video_tokenizers.networks import TokenizerConfigs, TokenizerModels

_DTYPE, _DEVICE = torch.bfloat16, "cuda"
_UINT8_MAX_F = float(torch.iinfo(torch.uint8).max)
_SPATIAL_ALIGN = 16
_TEMPORAL_ALIGN = 8


def load_jit_model(jit_filepath: str = None, device: str = "cuda") -> torch.jit.ScriptModule:
    """Loads a torch.jit.ScriptModule from a filepath.

    Args:
        jit_filepath: The filepath to the JIT-compiled model.
        device: The device to load the model onto, default=cuda.
    Returns:
        The JIT compiled model loaded to device and on eval mode.
    """
    model = torch.jit.load(jit_filepath)
    return model.eval().to(device)


def save_jit_model(
    model: torch.jit.ScriptModule | torch.jit.RecursiveScriptModule = None,
    jit_filepath: str = None,
) -> None:
    """Saves a torch.jit.ScriptModule or torch.jit.RecursiveScriptModule to file.

    Args:
        model: JIT compiled model loaded onto `config.checkpoint.jit.device`.
        jit_filepath: The filepath to the JIT-compiled model.
    """
    torch.jit.save(model, jit_filepath)


def get_filepaths(input_pattern) -> list[str]:
    """Returns a list of filepaths from a pattern."""
    filepaths = sorted(glob(str(input_pattern)))
    return list(set(filepaths))


def get_output_filepath(filepath: str, output_dir: str = None) -> str:
    """Returns the output filepath for the given input filepath."""
    output_dir = output_dir or f"{os.path.dirname(filepath)}/reconstructions"
    output_filepath = f"{output_dir}/{os.path.basename(filepath)}"
    os.makedirs(output_dir, exist_ok=True)
    return output_filepath


def read_image(filepath: str) -> np.ndarray:
    """Reads an image from a filepath.

    Args:
        filepath: The filepath to the image.

    Returns:
        The image as a numpy array, layout HxWxC, range [0..255], uint8 dtype.
    """
    image = media.read_image(filepath)
    # convert the grey scale image to RGB
    # since our tokenizers always assume 3-channel RGB image
    if image.ndim == 2:
        image = np.stack([image] * 3, axis=-1)
    # convert RGBA to RGB
    if image.shape[-1] == 4:
        image = image[..., :3]
    return image


def read_video(filepath: str) -> np.ndarray:
    """Reads a video from a filepath.

    Args:
        filepath: The filepath to the video.
    Returns:
        The video as a numpy array, layout TxHxWxC, range [0..255], uint8 dtype.
    """
    video = media.read_video(filepath)
    # convert the grey scale frame to RGB
    # since our tokenizers always assume 3-channel video
    if video.ndim == 3:
        video = np.stack([video] * 3, axis=-1)
    # convert RGBA to RGB
    if video.shape[-1] == 4:
        video = video[..., :3]
    return video


def resize_image(image: np.ndarray, short_size: int = None) -> np.ndarray:
    """Resizes an image to have the short side of `short_size`.

    Args:
        image: The image to resize, layout HxWxC, of any range.
        short_size: The size of the short side.
    Returns:
        The resized image.
    """
    if short_size is None:
        return image
    height, width = image.shape[-3:-1]
    if height <= width:
        height_new, width_new = short_size, int(width * short_size / height + 0.5)
        width_new = width_new if width_new % 2 == 0 else width_new + 1
    else:
        height_new, width_new = (
            int(height * short_size / width + 0.5),
            short_size,
        )
        height_new = height_new if height_new % 2 == 0 else height_new + 1
    return media.resize_image(image, shape=(height_new, width_new))


def resize_video(video: np.ndarray, short_size: int = None) -> np.ndarray:
    """Resizes a video to have the short side of `short_size`.

    Args:
        video: The video to resize, layout TxHxWxC, of any range.
        short_size: The size of the short side.
    Returns:
        The resized video.
    """
    if short_size is None:
        return video
    height, width = video.shape[-3:-1]
    if height <= width:
        height_new, width_new = short_size, int(width * short_size / height + 0.5)
        width_new = width_new if width_new % 2 == 0 else width_new + 1
    else:
        height_new, width_new = (
            int(height * short_size / width + 0.5),
            short_size,
        )
        height_new = height_new if height_new % 2 == 0 else height_new + 1
    return media.resize_video(video, shape=(height_new, width_new))


def write_image(filepath: str, image: np.ndarray):
    """Writes an image to a filepath."""
    return media.write_image(filepath, image)


def write_video(filepath: str, video: np.ndarray, fps: int = 24) -> None:
    """Writes a video to a filepath."""
    return media.write_video(filepath, video, fps=fps)


def numpy2tensor(
    input_image: np.ndarray,
    dtype: torch.dtype = _DTYPE,
    device: str = _DEVICE,
    range_min: int = -1,
) -> torch.Tensor:
    """Converts image(dtype=np.uint8) to `dtype` in range [0..255].

    Args:
        input_image: A batch of images in range [0..255], BxHxWx3 layout.
    Returns:
        A torch.Tensor of layout Bx3xHxW in range [-1..1], dtype.
    """
    ndim = input_image.ndim
    indices = list(range(1, ndim))[-1:] + list(range(1, ndim))[:-1]
    image = input_image.transpose((0,) + tuple(indices)) / _UINT8_MAX_F
    if range_min == -1:
        image = 2.0 * image - 1.0
    return torch.from_numpy(image).to(dtype).to(device)


def tensor2numpy(input_tensor: torch.Tensor, range_min: int = -1) -> np.ndarray:
    """Converts tensor in [-1,1] to image(dtype=np.uint8) in range [0..255].

    Args:
        input_tensor: Input image tensor of Bx3xHxW layout, range [-1..1].
    Returns:
        A numpy image of layout BxHxWx3, range [0..255], uint8 dtype.
    """
    if range_min == -1:
        input_tensor = (input_tensor.float() + 1.0) / 2.0
    ndim = input_tensor.ndim
    output_image = input_tensor.clamp(0, 1).cpu().numpy()
    output_image = output_image.transpose((0,) + tuple(range(2, ndim)) + (1,))
    return (output_image * _UINT8_MAX_F + 0.5).astype(np.uint8)


def pad_image_batch(batch: np.ndarray, spatial_align: int = _SPATIAL_ALIGN) -> tuple[np.ndarray, list[int]]:
    """Pads a batch of images to be divisible by `spatial_align`.

    Args:
        batch: The batch of images to pad, layout BxHxWx3, in any range.
        align: The alignment to pad to.
    Returns:
        The padded batch and the crop region.
    """
    height, width = batch.shape[1:3]
    align = spatial_align
    height_to_pad = (align - height % align) if height % align != 0 else 0
    width_to_pad = (align - width % align) if width % align != 0 else 0

    crop_region = [
        height_to_pad >> 1,
        width_to_pad >> 1,
        height + (height_to_pad >> 1),
        width + (width_to_pad >> 1),
    ]
    batch = np.pad(
        batch,
        (
            (0, 0),
            (height_to_pad >> 1, height_to_pad - (height_to_pad >> 1)),
            (width_to_pad >> 1, width_to_pad - (width_to_pad >> 1)),
            (0, 0),
        ),
        mode="constant",
    )
    return batch, crop_region


def pad_video_batch(
    batch: np.ndarray,
    temporal_align: int = _TEMPORAL_ALIGN,
    spatial_align: int = _SPATIAL_ALIGN,
) -> tuple[np.ndarray, list[int]]:
    """Pads a batch of videos to be divisible by `temporal_align` or `spatial_align`.

    Zero pad spatially. Reflection pad temporally to handle causality better.
    Args:
        batch: The batch of videos to pad., layout BxFxHxWx3, in any range.
        align: The alignment to pad to.
    Returns:
        The padded batch and the crop region.
    """
    num_frames, height, width = batch.shape[-4:-1]
    align = spatial_align
    height_to_pad = (align - height % align) if height % align != 0 else 0
    width_to_pad = (align - width % align) if width % align != 0 else 0

    align = temporal_align
    frames_to_pad = (align - (num_frames - 1) % align) if (num_frames - 1) % align != 0 else 0

    crop_region = [
        frames_to_pad >> 1,
        height_to_pad >> 1,
        width_to_pad >> 1,
        num_frames + (frames_to_pad >> 1),
        height + (height_to_pad >> 1),
        width + (width_to_pad >> 1),
    ]
    batch = np.pad(
        batch,
        (
            (0, 0),
            (0, 0),
            (height_to_pad >> 1, height_to_pad - (height_to_pad >> 1)),
            (width_to_pad >> 1, width_to_pad - (width_to_pad >> 1)),
            (0, 0),
        ),
        mode="constant",
    )
    batch = np.pad(
        batch,
        (
            (0, 0),
            (frames_to_pad >> 1, frames_to_pad - (frames_to_pad >> 1)),
            (0, 0),
            (0, 0),
            (0, 0),
        ),
        mode="edge",
    )
    return batch, crop_region


def unpad_video_batch(batch: np.ndarray, crop_region: list[int]) -> np.ndarray:
    """Unpads video with `crop_region`.

    Args:
        batch: A batch of numpy videos, layout BxFxHxWxC.
        crop_region: [f1,y1,x1,f2,y2,x2] first, top, left, last, bot, right crop indices.

    Returns:
        np.ndarray: Cropped numpy video, layout BxFxHxWxC.
    """
    assert len(crop_region) == 6, "crop_region should be len of 6."
    f1, y1, x1, f2, y2, x2 = crop_region
    return batch[..., f1:f2, y1:y2, x1:x2, :]


def unpad_image_batch(batch: np.ndarray, crop_region: list[int]) -> np.ndarray:
    """Unpads image with `crop_region`.

    Args:
        batch: A batch of numpy images, layout BxHxWxC.
        crop_region: [y1,x1,y2,x2] top, left, bot, right crop indices.

    Returns:
        np.ndarray: Cropped numpy image, layout BxHxWxC.
    """
    assert len(crop_region) == 4, "crop_region should be len of 4."
    y1, x1, y2, x2 = crop_region
    return batch[..., y1:y2, x1:x2, :]


def get_pytorch_model(jit_filepath: str = None, tokenizer_config: str = None):
    tokenizer_name = tokenizer_config["name"]
    model = TokenizerModels[tokenizer_name].value(**tokenizer_config)
    ckpts = torch.jit.load(jit_filepath)
    return model, ckpts


def load_pytorch_model(jit_filepath: str, tokenizer_config: dict, model_type: str, device):
    """Loads a torch.nn.Module from a filepath."""
    model, ckpts = get_pytorch_model(jit_filepath, tokenizer_config)
    if model_type == "enc":
        model = model.encoder_jit()
    elif model_type == "dec":
        model = model.decoder_jit()
    model.load_state_dict(ckpts.state_dict(), strict=False)
    return model.eval().to(tokenizer_config["dtype"]).to(device)


def get_tokenizer_config(tokenizer_type) -> TokenizerConfigs:
    """return tokeinzer config from tokenizer name"""
    match = re.match(r"Cosmos-Tokenizer-(\D+)(\d+)x(\d+).*", tokenizer_type)
    if match:
        name, temporal, spatial = match.groups()
        tokenizer_config = TokenizerConfigs[name].value
        tokenizer_config.update(dict(spatial_compression=int(spatial)))
        tokenizer_config.update(dict(temporal_compression=int(temporal)))
        return tokenizer_config
    return None



# Copyright (c) 2025, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import numpy as np
import torch
import wrapt
from transformers import AutoModel, AutoTokenizer

from nemo.deploy import ITritonDeployable
from nemo.export.utils import get_example_inputs, get_model_device_type, is_nemo2_checkpoint, validate_fp8_network
from nemo.utils import logging

if TYPE_CHECKING:
    import tensorrt as trt


@wrapt.decorator
def noop_decorator(func):
    """No op decorator"""

    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


use_pytriton = True
batch = noop_decorator
try:
    from pytriton.decorators import batch
except Exception:
    logging.warning("PyTriton is not available.")
    use_pytriton = False


use_onnxruntime = True
try:
    import onnxruntime
except Exception:
    logging.warning("onnxruntime is not available.")
    use_onnxruntime = False


use_trt = True
try:
    import tensorrt as trt
except ImportError:
    logging.warning("tensorrt is not available")
    use_trt = False


# pylint: disable=line-too-long
class OnnxLLMExporter(ITritonDeployable):
    """
    Exports models to ONNX and run fast inference.

    Example:
        from nemo.export.onnx_llm_exporter import OnnxLLMExporter

        onnx_llm_exporter = OnnxLLMExporter(
            onnx_model_dir="/path/for/onnx_model/files",
            model_name_or_path="/path/for/model/files",
        )

        onnx_llm_exporter.export(
            input_names=["input_ids", "attention_mask", "dimensions"],
            output_names=["embeddings"],
        )

        output = onnx_llm_exporter.forward(["Hi, how are you?", "I am good, thanks, how about you?"])
        print("output: ", output)
    """

    def __init__(
        self,
        onnx_model_dir: str,
        model: Optional[torch.nn.Module] = None,
        tokenizer=None,
        model_name_or_path: str = None,
        load_runtime: bool = True,
    ):
        """
        Initializes the ONNX Exporter.

        Args:
            onnx_model_dir (str): path for storing the ONNX model files.
            model (Optional[torch.nn.Module]): torch model.
            tokenizer (HF or NeMo tokenizer): tokenizer class.
            model_name_or_path (str): a path for ckpt or HF model ID
            load_runtime (bool): load ONNX runtime if there is any exported model available in
                                 the onnx_model_dir folder.
        """
        self.onnx_model_dir = onnx_model_dir
        self.model_name_or_path = model_name_or_path
        self.onnx_model_path = str(Path(onnx_model_dir) / "model.onnx")
        self.model = model
        self.tokenizer = tokenizer
        self.model_input_names = None
        self.model_output_names = None
        self.onnx_runtime_session = None
        self.calibration_data = None
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.quant_max_batch_size = None

        if self.model_name_or_path is not None:
            if model is not None:
                raise ValueError("A model was also passed but it will be overridden.")

            if Path(self.model_name_or_path).is_dir():
                if is_nemo2_checkpoint(self.model_name_or_path):
                    raise NotImplementedError("NeMo 2.0 checkpoint will be supported later.")
                else:
                    self._load_hf_model()

        if load_runtime:
            self._load_runtime()

    def _load_runtime(self):
        if use_onnxruntime:
            if Path(self.onnx_model_path).exists():
                self.onnx_runtime_session = onnxruntime.InferenceSession(self.onnx_model_path)
                self.model_input_names = [input.name for input in self.onnx_runtime_session.get_inputs()]
                self.model_output_names = [output.name for output in self.onnx_runtime_session.get_outputs()]
                self.tokenizer = AutoTokenizer.from_pretrained(
                    Path(self.onnx_model_dir) / "tokenizer", trust_remote_code=True
                )

    def _load_hf_model(self):
        self.model = AutoModel.from_pretrained(
            self.model_name_or_path,
            trust_remote_code=True,
        ).eval()

        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name_or_path, trust_remote_code=True)

    def export(
        self,
        input_names: list,
        output_names: list,
        example_inputs: dict = None,
        opset: int = 20,
        dynamic_axes_input: Optional[dict] = None,
        dynamic_axes_output: Optional[dict] = None,
        export_dtype: str = "fp32",
        verbose: bool = False,
    ):
        """
        Performs ONNX conversion from a PyTorch model.

        Args:
            input_names (list): input parameter names of the model that ONNX will export will use.
            output_names (list): output parameter names of the model that ONNX will export will use.
            example_inputs (dict): example input for the model to build the engine.
            opset (int): ONNX opset version. Default is 20.
            dynamic_axes_input (dict): Variable length axes for the input.
            dynamic_axes_output (dict): Variable length axes for the output.
            export_dtype (str): Export dtype, fp16 or fp32.
            verbose (bool): Enable verbose or not.
        """

        self._export_to_onnx(
            input_names=input_names,
            example_inputs=example_inputs,
            output_names=output_names,
            opset=opset,
            dynamic_axes_input=dynamic_axes_input,
            dynamic_axes_output=dynamic_axes_output,
            export_dtype=export_dtype,
            verbose=verbose,
        )
        self._load_runtime()

    def _export_to_onnx(
        self,
        input_names: list,
        output_names: list,
        example_inputs: dict = None,
        opset: int = 20,
        dynamic_axes_input: Optional[dict] = None,
        dynamic_axes_output: Optional[dict] = None,
        export_dtype: Union[torch.dtype, str] = "fp32",
        verbose: bool = False,
    ):

        if example_inputs is None:
            example_inputs = get_example_inputs(self.tokenizer)

        if "dimensions" in input_names:
            example_inputs["dimensions"] = torch.tensor([1] * example_inputs["input_ids"].shape[0])

        if isinstance(export_dtype, str):
            export_dtype = {"fp16": torch.float16, "fp32": torch.float32}[export_dtype]

        self.model.to(export_dtype)

        Path(self.onnx_model_dir).mkdir(parents=True, exist_ok=True)

        with torch.autocast(device_type=get_model_device_type(self.model), dtype=export_dtype):
            torch.onnx.export(
                model=self.model,
                args=(example_inputs,),
                f=self.onnx_model_path,
                input_names=input_names,
                output_names=output_names,
                dynamic_axes={**dynamic_axes_input, **dynamic_axes_output},
                verbose=verbose,
                opset_version=opset,
            )
        logging.info(f"Successfully exported PyTorch model to ONNX model {self.onnx_model_path}")

        existing_directory_path = Path(self.onnx_model_dir) / "tokenizer"
        existing_directory_path.mkdir(exist_ok=True)
        self.tokenizer.save_pretrained(existing_directory_path)

    def export_onnx_to_trt(
        self,
        trt_model_dir: str,
        profiles=None,
        override_layernorm_precision_to_fp32: bool = False,
        override_layers_to_fp32: List = None,
        trt_dtype: str = "fp16",
        profiling_verbosity: str = "layer_names_only",
        trt_builder_flags: List["trt.BuilderFlag"] = None,
    ) -> None:
        """Performs TensorRT conversion from an ONNX model.

        Args:
            trt_model_dir: path to store the TensorRT model.
            profiles: TensorRT profiles.
            override_layernorm_precision_to_fp32 (bool): whether to convert layers to fp32 or not.
            override_layers_to_fp32 (List): Layer names to be converted to fp32.
            trt_dtype (str): "fp16" or "fp32".
            profiling_verbosity (str): Profiling verbosity. Default is "layer_names_only".
            trt_builder_flags (List[trt.BuilderFlag]): TRT specific flags.
        """
        logging.info(f"Building TRT engine from ONNX model ({self.onnx_model_path})")
        trt_logger = trt.Logger(trt.Logger.WARNING)
        builder = trt.Builder(trt_logger)
        network = builder.create_network(1 << int(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH))
        config = builder.create_builder_config()
        parser = trt.OnnxParser(network, trt_logger)

        # we use parse_from_file() instead of parse() because it can be used for both single
        # file models as well as externally stored models (required when model >2GiB)
        if not parser.parse_from_file(self.onnx_model_path):
            logging.warning("ONNX model could not be parsed")
            for error in range(parser.num_errors):
                logging.error(parser.get_error(error))
            return

        if profiles:
            for profile in profiles:
                optimization_profile = builder.create_optimization_profile()

                for i in range(network.num_inputs):
                    in_tensor = network.get_input(i)
                    optimization_profile.set_shape(
                        in_tensor.name,
                        min=profile[in_tensor.name][0],
                        opt=profile[in_tensor.name][1],
                        max=profile[in_tensor.name][2],
                    )

                config.add_optimization_profile(optimization_profile)

        if trt_dtype == "fp16":
            logging.info("Setting Build Flag FP16")
            config.set_flag(trt.BuilderFlag.FP16)
        elif trt_dtype == "fp8":
            # With FP8 export we want to also enable FP16 layers as a fallback instead of FP32
            logging.info("Setting Build Flag FP8 and FP16")
            config.set_flag(trt.BuilderFlag.FP8)
            config.set_flag(trt.BuilderFlag.FP16)
            validate_fp8_network(network)

        # patch network
        if override_layernorm_precision_to_fp32:
            logging.info("Overriding TensorRT network LayerNorm precision to float32.")
            self._override_layernorm_precision_to_fp32(network)

        if override_layers_to_fp32:
            logging.info("Overriding some layers to float32.")
            self._override_layers_to_fp32(network, override_layers_to_fp32)

        try:
            config.profiling_verbosity = {
                "detailed": trt.ProfilingVerbosity.DETAILED,
                "layer_names_only": trt.ProfilingVerbosity.LAYER_NAMES_ONLY,
                "none": trt.ProfilingVerbosity.NONE,
            }[profiling_verbosity]
        except KeyError:
            error_msg = "Unknown profiling verbosity value."
            raise ValueError(error_msg)
        logging.info(f"Setting Profiling Verbosity to {config.profiling_verbosity}")

        if trt_builder_flags is not None:
            for flag in trt_builder_flags:
                config.set_flag(flag)

        engine_string = builder.build_serialized_network(network, config)
        if engine_string is None:
            raise Exception("Failed to serialize the TensorRT Engine. Please check the " "TensorRT logs for details")

        trt_model_path = Path(trt_model_dir)
        trt_model_path.mkdir(parents=True, exist_ok=True)
        trt_model_path = trt_model_path / "model.plan"
        trt_model_path.write_bytes(engine_string)
        logging.info(f"Successfully exported ONNX model ({self.onnx_model_path}) " f"to TRT engine ({trt_model_path})")

    def _override_layer_precision_to_fp32(self, layer: "trt.ILayer") -> None:
        layer.precision = trt.float32
        layer.set_output_type(0, trt.float32)

    def _override_layers_to_fp32(self, network: "trt.INetworkDefinition", fp32_layer_patterns: list[str]) -> None:
        for i in range(network.num_layers):
            layer = network.get_layer(i)
            layer_name = layer.name
            if any(layer_name.startswith(pattern) for pattern in fp32_layer_patterns) and layer.precision in {
                trt.float32,
                trt.float16,
            }:
                if layer.type in {trt.LayerType.CAST}:
                    logging.info(f"Skipping overriding {layer.type} layer {i} {layer_name} dtype")
                    continue
                if any(
                    layer.get_input(input_idx).dtype in {trt.float32, trt.float16}
                    for input_idx in range(layer.num_inputs)
                ):
                    # Note: Assigning to layer.precision (even the same value) sets precision_is_set=True,
                    # which prevents TensorRT from changing this layer's precision
                    layer.precision = trt.float32
                    logging.info(f"Setting layer {i} {layer_name} (type: {layer.type}) precision to FP32")
                for j in range(layer.num_outputs):
                    if layer.get_output_type(j) in {trt.float32, trt.float16}:
                        layer.set_output_type(j, trt.float32)
                        logging.info(f"Setting layer {i} {layer_name} (type: {layer.type}) output type {j} to FP32")

    def _override_layernorm_precision_to_fp32(self, network: "trt.INetworkDefinition") -> None:
        """Set the precision of LayerNorm subgraphs to FP32 to preserve accuracy.

        - https://nvbugs/4478448 (Mistral)
        - https://nvbugs/3802112 (T5)

        Args:
            network: tensorrt.INetworkDefinition
        """
        # Logic originally from OSS T5 HF export script:
        # https://gitlab-master.nvidia.com/TensorRT/Public/oss/-/blob/77495ec/demo/HuggingFace/T5/export.py
        pow_ops = {}
        for layer_index, layer in enumerate(network):
            if layer.type == trt.LayerType.IDENTITY:
                all_fp32 = all(
                    [
                        layer.output_type_is_set(o) and layer.get_output_type(o) == trt.float32
                        for o in range(layer.num_outputs)
                    ]
                )
                if all_fp32:
                    if layer.get_input(0).dtype == trt.float32:
                        layer.precision = trt.float32

            if layer.type == trt.LayerType.ELEMENTWISE:
                layer.__class__ = getattr(trt, "IElementWiseLayer")
                if layer.op == trt.ElementWiseOperation.POW:
                    pow_ops[layer] = layer_index
                    self._override_layer_precision_to_fp32(layer)

        for _, index in pow_ops.items():
            # Iterate from few layers before pow to include residual add and cast op.
            # Iterate till 10 layers after pow op to include all
            # operations included in layer norm.
            START_OFFSET = 4
            END_OFFSET = 12
            for i in range(index - START_OFFSET, index + END_OFFSET):
                layer = network.get_layer(i)
                if layer.type == trt.LayerType.REDUCE:
                    self._override_layer_precision_to_fp32(layer)

                if layer.type == trt.LayerType.ELEMENTWISE:
                    layer.__class__ = getattr(trt, "IElementWiseLayer")
                    if layer.op == trt.ElementWiseOperation.SUM:
                        self._override_layer_precision_to_fp32(layer)

                if layer.type == trt.LayerType.UNARY:
                    layer.__class__ = getattr(trt, "IUnaryLayer")
                    if layer.op == trt.UnaryOperation.SQRT:
                        self._override_layer_precision_to_fp32(layer)

                if layer.type == trt.LayerType.ELEMENTWISE:
                    layer.__class__ = getattr(trt, "IElementWiseLayer")
                    if layer.op == trt.ElementWiseOperation.DIV:
                        self._override_layer_precision_to_fp32(layer)

                if layer.type == trt.LayerType.ELEMENTWISE:
                    layer.__class__ = getattr(trt, "IElementWiseLayer")
                    if layer.op == trt.ElementWiseOperation.PROD:
                        self._override_layer_precision_to_fp32(layer)

    def forward(self, inputs: Union[List, Dict], dimensions: Optional[List] = None):
        """Run inference for a given input.

        Args:
            inputs (Union[List, Dict]): Input for the model. If list, it should be a list of strings.
                If dict, it should be a dictionary with keys as the model input names.
            dimensions (Optional[List]): The dimensions parameter of the model. Required if the model
                was exported to accept dimensions parameter and inputs is given as a list of strings.

        Returns:
            np.ndarray: Model output.
        """

        if self.onnx_runtime_session is None:
            warnings.warn("ONNX Runtime is not available. Please install the onnxruntime-gpu and try again.")
            return None

        if isinstance(inputs, List):
            if "dimensions" in self.model_input_names and dimensions is None:
                raise ValueError("Dimensions should be provided for list input.")
            inputs = dict(self.tokenizer(inputs))
            inputs["dimensions"] = dimensions

        output = self.onnx_runtime_session.run(self.model_output_names, inputs)
        return output[0]

    @property
    def get_model(self):
        """Returns the model"""

        return self.model

    @property
    def get_tokenizer(self):
        """Returns the tokenizer"""

        return self.tokenizer

    @property
    def get_model_input_names(self):
        """Returns the model input names"""

        return self.model_input_names

    @property
    def get_triton_input(self):
        """Get triton input"""

        raise NotImplementedError("This function will be implemented later.")

    @property
    def get_triton_output(self):
        """Get triton output"""

        raise NotImplementedError("This function will be implemented later.")

    @batch
    def triton_infer_fn(self, **inputs: np.ndarray):
        """PyTriton inference function"""

        raise NotImplementedError("This function will be implemented later.")
